#!/usr/bin/env python
#
# test_icd10.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import textwrap as tw

from unittest import mock

import pytest

import numpy  as np
import pandas as pd

import funpack.icd10     as icd10
import funpack.hierarchy as hierarchy

from . import tempdir

def test_store_saveCodes():

    with tempdir():

        icd10.storeCodes.store = []

        with pytest.raises(ValueError):
            icd10.saveCodes('file', None, ['badfield'])

        codings = tw.dedent("""
        coding\tmeaning\tnode_id\tparent_id
        a10\ta desc\t5\t0
        b20\tb desc\t1\t5
        c30\tc desc\t3\t5
        d40\td desc\t4\t3
        e50\te desc\t2\t1
        f60\tf desc\t9\t9
        """)

        with open('codings.tsv', 'wt') as f:
            f.write(codings)

        codes = ['a10', np.nan, 'b20', 'c30', 'd40', 'e50']

        icd10.storeCodes(codes[:2])
        icd10.storeCodes(codes[2:])

        del codes[1]

        h = hierarchy.loadHierarchyFile('codings.tsv')

        with mock.patch('funpack.hierarchy.getHierarchyFilePath',
                        return_value='codings.tsv'):
            icd10.saveCodes(
                'mapping.tsv', h,
                fields=['code', 'value', 'description',
                        'parent_codes', 'parent_descs'])

        values = [h.index(c) + 1 for c in codes]

        mf      = pd.read_csv('mapping.tsv', delimiter='\t', index_col=False)
        descs   = ['a desc', 'b desc', 'c desc', 'd desc', 'e desc']
        pcodes  = [np.nan, 'a10', 'a10', 'a10,c30', 'a10,b20']
        pdescs  = [np.nan, '[a desc]', '[a desc]',
                   '[a desc] [c desc]', '[a desc] [b desc]']

        gotpcodes = mf['parent_codes']
        gotpdescs = mf['parent_descs']

        assert (mf['code']         == codes) .all()
        assert (mf['value']        == values).all()
        assert (mf['description']  == descs) .all()
        assert np.isnan(gotpcodes.iloc[0])
        assert np.isnan(gotpdescs.iloc[0])
        assert (gotpcodes.iloc[1:] == pcodes[1:]).all()
        assert (gotpdescs.iloc[1:] == pdescs[1:]).all()



#!/usr/bin/env python
#
# config.py - Parses command line arguments and configuration files.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions for parsing ``funpack`` command line
arguments and configuration files.
"""


import os.path         as op
import functools       as ft
import itertools       as it
import multiprocessing as mp
import                    sys
import                    shlex
import                    logging
import                    argparse
import                    collections
import numpy           as np

import funpack
import funpack.util           as util
import funpack.custom         as custom
import funpack.importing      as importing
import funpack.exporting      as exporting
import funpack.exporting_hdf5 as exporting_hdf5
import funpack.exporting_tsv  as exporting_tsv


log = logging.getLogger(__name__)


VERSION                    = funpack.__version__
DEFAULT_MERGE_AXIS         = importing.MERGE_AXIS
DEFAULT_MERGE_STRATEGY     = importing.MERGE_STRATEGY
DEFAULT_EXPORT_FORMAT      = exporting.EXPORT_FORMAT
AVAILABLE_MERGE_AXES       = importing.MERGE_AXIS_OPTIONS
AVAILABLE_MERGE_STRATEGIES = importing.MERGE_STRATEGY_OPTIONS
DEFAULT_COLUMN_PATTERN     = exporting.COLUMN_PATTERN
DEFAULT_TSV_SEP            = exporting_tsv.TSV_SEP
DEFAULT_HDF5_KEY           = exporting_hdf5.HDF5_KEY
DEFAULT_HDF5_STYLE         = exporting_hdf5.HDF5_STYLE
AVAILABLE_HDF5_STYLES      = exporting_hdf5.HDF5_STYLES
CLI_ARGUMENTS              = collections.OrderedDict((

    ('Inputs', [
        ((       'outfile',),        {}),
        ((       'infile',),         {'nargs'   : '+'}),
        (('e',   'encoding'),        {'action'  : 'append'}),
        (('l',   'loader'),          {'nargs'   : 2,
                                      'metavar' : ('FILE', 'LOADER'),
                                      'action'  : 'append'}),
        (('i',   'index'),           {'nargs'   : 2,
                                      'metavar' : ('FILE', 'INDEX'),
                                      'action'  : 'append'}),
        (('ma',  'merge_axis'),      {'choices' : AVAILABLE_MERGE_AXES,
                                      'default' : DEFAULT_MERGE_AXIS}),
        (('ms',  'merge_strategy'),  {'choices' : AVAILABLE_MERGE_STRATEGIES,
                                      'default' : DEFAULT_MERGE_STRATEGY}),
        (('cfg', 'config_file'),     {}),
        (('vf',  'variable_file'),   {'action'  : 'append'}),
        (('df',  'datacoding_file'), {'action'  : 'append'}),
        (('tf',  'type_file'),       {}),
        (('pf',  'processing_file'), {}),
        (('cf',  'category_file'),   {})]),

    ('Import options', [
        (('ia', 'import_all'),     {'action' : 'store_true'}),
        (('r',  'remove_unknown'), {'action' : 'store_true'}),
        (('tt', 'trust_types'),    {'action' : 'store_true'}),
        (('s',  'subject'),        {'action' : 'append'}),
        (('v',  'variable'),       {'action' : 'append'}),
        (('co', 'column'),         {'action' : 'append'}),
        (('c',  'category'),       {'action' : 'append'}),
        (('vi', 'visit'),          {'action' : 'append'}),
        (('ex', 'exclude'),        {'action' : 'append'})]),

    ('Cleaning options', [
        (('sn',  'skip_insertna'),      {'action'  : 'store_true'}),
        (('scv', 'skip_childvalues'),   {'action'  : 'store_true'}),
        (('scf', 'skip_clean_funcs'),   {'action'  : 'store_true'}),
        (('sr',  'skip_recoding'),      {'action'  : 'store_true'}),
        (('nv',  'na_values'),          {'nargs'   : 2,
                                         'action'  : 'append',
                                         'metavar' : ('VID', 'NAVALUES')}),
        (('re',  'recoding'),           {'nargs'   : 3,
                                         'action'  : 'append',
                                         'metavar' : ('VID',
                                                      'RAWLEVELS',
                                                      'NEWLEVELS')}),
        (('cv',  'child_values'),       {'nargs'   : 3,
                                         'action'  : 'append',
                                         'metavar' : ('VID',
                                                      'EXPRS',
                                                      'VALUES')}),
        (('cl',  'clean'),              {'nargs'   : 2,
                                         'action'  : 'append',
                                         'metavar' : ('VID', 'EXPR')}),
        (('tc',  'type_clean'),         {'nargs'   : 2,
                                         'action'  : 'append',
                                         'metavar' : ('TYPE', 'EXPR')}),
        (('gc',  'global_clean'),       {'metavar' : 'EXPR'})]),

    ('Processing options', [
        (('sp',  'skip_processing'), {'action'  : 'store_true'}),
        (('ppr', 'prepend_process'), {'action'  : 'append',
                                      'nargs'   : 2,
                                      'metavar' : ('VARS', 'PROCS')}),
        (('apr', 'append_process'),  {'action'  : 'append',
                                      'nargs'   : 2,
                                      'metavar' : ('VARS', 'PROCS')})]),

    ('Export options', [
        (('f',    'format'),            {'default' : DEFAULT_EXPORT_FORMAT}),
        (('cp',   'column_pattern'),    {'default' : DEFAULT_COLUMN_PATTERN}),
        (('rc',   'rename_column'),     {'action'  : 'append',
                                         'nargs'   : 2,
                                         'metavar' : ('OLD_NAME', 'NEW_NAME')}),  # noqa
        (('oi',   'output_id_column'),  {}),
        (('edf',  'date_format'),       {'default' : 'default'}),
        (('etf',  'time_format'),       {'default' : 'default'}),
        (('nr',   'num_rows'),          {'type'    : int}),
        (('uf',   'unknown_vars_file'), {}),
        (('imf',  'icd10_map_file'),    {}),
        (('def',  'description_file'),  {}),
        (('sf',   'summary_file'),      {})]),

    ('TSV export options', [
        (('ts',  'tsv_sep'),            {'default' : DEFAULT_TSV_SEP}),
        (('tm',  'tsv_missing_values'), {'default' : ''}),
        (('nn',  'non_numeric_file'),   {}),
        (('tvf', 'tsv_var_format'),     {'nargs'   : 2,
                                         'metavar' : ('VID', 'FORMATTER'),
                                         'action'  : 'append'})]),

    ('HDF5 export options', [

        (('hk', 'hdf5_key'),   {'default' : DEFAULT_HDF5_KEY}),
        (('hs', 'hdf5_style'), {'default' : DEFAULT_HDF5_STYLE,
                                'choices' : AVAILABLE_HDF5_STYLES})]),

    ('Miscellaneous options', [
        (('V',  'version'),      {'action'  : 'version',
                                  'version' :
                                  '%(prog)s {}'.format(VERSION)}),
        (('d',  'dry_run'),      {'action' : 'store_true'}),
        (('nb', 'no_builtins'),  {'action' : 'store_true'}),
        (('lm', 'low_memory'),   {'action' : 'store_true'}),
        (('wd', 'work_dir'),     {}),
        (('lf', 'log_file'),     {}),
        (('nj', 'num_jobs'),     {'type'    : int,
                                  'default' : mp.cpu_count()}),
        (('pt', 'pass_through'), {'action'  : 'store_true'}),
        (('p',  'plugin_file'),  {'action'  : 'append',
                                  'metavar' : 'FILE'}),
        (('ow', 'overwrite'),    {'action'  : 'store_true'}),
        (('n',  'noisy'),        {'action'  : 'count'}),
        (('q',  'quiet'),        {'action'  : 'store_true'})])))


CLI_DESCRIPTIONS = {

    'Inputs' :
    'The --variable_file and --datacoding_file options can be used multiple\n'
    'times - all provided files will be merged into a single table using the\n'
    'variable/data coding IDs.',

    'Import options' :
    'Using the --import_all option will increase RAM and CPU requirements,\n'
    'as it forces every column to be loaded and processed. The purpose of\n'
    'this option is to allow the user to identify previously unknown\n'
    'columns that passed the processing steps (e.g. the sparsity test), and\n'
    'therefore may need to be looked at more closely. Intended to be used\n'
    'with the --unknown_vars_file export option.',

    'Export options' :
    'The --unknown_vars_file (only active if --import_all is also active)\n'
    'allows a file to be saved containing information about columns which\n'
    'were in the input data, but were either not in the variable table, or\n'
    'were uncategorised and did not have any cleaning/processing rules\n'
    'specified. It contains five columns - the column name, the originating\n'
    'input file, the reason the column is being included (either unknown\n'
    'or uncategorised/unprocessed), whether the column passed the processing\n'
    'stage (e.g. sparsity/redundancy checks), and whether the column was\n'
    'exported.',
}


CLI_ARGUMENT_HELP = {

    # Input options
    'outfile' : 'Location to store output data.',
    'infile'  : 'File(s) containing input data.',

    'encoding' :
    'Character encoding. A single encoding can be specified, or this option '
    'can be used multiple times, one for each input file.',

    'loader' :
    'Use custom loader for file. Can be used multiple times.',

    'index' :
    'Position of index column for file (starting from 0). Can be used '
    'multiple times. Defaults to 0.',

    'merge_axis' :
    'Axis to concatenate multiple input files along (default: "{}"). '
    'Options are "subjects"/"rows"/"0" or "variables"/"columns"/'
    '"1".'.format(DEFAULT_MERGE_AXIS),

    'merge_strategy' :
    'Strategy for merging multiple input files (default: "{}"). '
    'Options are "naive", "intersection"/"inner", or "union"/'
    '"outer".'.format(DEFAULT_MERGE_STRATEGY),

    'config_file' :
    'File containing default command line arguments.',

    'variable_file' :
    'File(s) containing rules for handling variables',

    'datacoding_file' :
    'File(s) containing rules for handling data codings.',

    'type_file' :
    'File containing rules for handling types.',

    'processing_file' :
    'File containing variable processing rules.',

    'category_file' :
    'File containing variable categories.',

    # Import options
    'import_all' :
    'Import and process all columns, and apply --variable/--category/'
    '--remove_unknown options after processing.',

    'remove_unknown' :
    'Remove variables which are not listed in variable table. Implied if '
    'variables are specified using --variable or --category.',

    'trust_types' :
    'Assume that columns in the input data with a known numeric type do not '
    'contain any bad/unparseable values. Using this flag will improve import '
    'performance, but will cause funpack to crash if the input file(s) do '
    'contain bad values.',

    'subject' :
    'Subject ID, range, comma-separated list, or file containing a list of '
    'subject IDs, or variable expression, denoting subjects to include. Can '
    'be used multiple times.',

    'variable' :
    'Variable ID, range, comma-separated list, or file containing a list of '
    'variable IDs, to import. Can be used multiple times. Implies '
    '--remove_unknown.',

    'column' :
    'Name of column to import, or file containing a list of column names. Can '
    'also be a glob-style wildcard pattern - columns with a name matching the '
    'pattern will be imported. Can be used multiple times. Implies '
    '--remove_unknown.',

    'category' :
    'Category ID or label to import. Can be used multiple times. Implies '
    '--remove_unknown.',

    'visit' :
    'Import this visit. Can be used multiple times. Allowable values are '
    '\'first\', \'last\', or a visit number.',

    'exclude' :
    'Subject ID, range, comma-separated list, or file containing a list of '
    'subject IDs specifying subjects to exclude. Can be used multiple times.',

    # Clean options
    'skip_insertna' :
    'Skip NA insertion.',
    'skip_childvalues' :
    'Skip child value replacement.',
    'skip_clean_funcs' :
    'Skip cleaning functions defined in variable table.',
    'skip_recoding' :
    'Skip raw->new level recoding.',

    'na_values' :
    'Replace these values with NA (overrides any NA values specified in '
    'variable table). Can be used multiple times. Values must be specified as '
    'a comma-separated list.',

    'recoding' :
    'Recode categorical variables (overrides any raw/new level recodings '
    'specified in variable table). Can be used multiple times. Raw and new '
    'level values must be specified as comma-separated lists.',

    'child_values' :
    'Replace NA with the given values where the given expressions evaluate '
    'to true (overrides any parent/child values specified in variable table). '
    'Can be used multiple times. Parent value expressions and corresponding '
    'child values must be specified as comma-separated lists.',

    'clean' :
    'Apply cleaning function(s) to variable (overrides any cleaning '
    'specified in variable table).',

    'type_clean' :
    'Apply clean function(s) to all variables of the specified type '
    '(overrides any cleaning specified in type table).',

    'global_clean' :
    'Apply cleaning function(s) to every variable (after variable-'
    'specific cleaning specified in variable table or via --clean).',

    # Processing options
    'skip_processing' :
    'Skip processing functions defined in processing table (but not '
    'those specified via --prepend_process and --append_process).',
    'prepend_process' :
    'Apply processing function(s) before processing defined in processing '
    'table.',
    'append_process' :
    'Apply processing function(s) after processing defined in processing '
    'table.',

    # Export options
    'format' :
    'Output file format (default: "{}").'.format(DEFAULT_EXPORT_FORMAT),

    'column_pattern' :
    'Pattern defining output column names (default: "{}").'.format(
        DEFAULT_COLUMN_PATTERN),

    'rename_column' :
    'Rename the given column instead of applying --column_pattern. Can '
    'be used multiple times',

    'output_id_column' :
    'Name of ID column in output file.',

    'date_format' :
    'Formatter to use for date variables (default: "default").',

    'time_format' :
    'Formatter to use for time variables (default: "default").',

    'num_rows' :
    'Number of rows to write at a time.',

    'unknown_vars_file' :
    'Save list of unknown variables/columns to file. Only applicable if '
    '--import_all is enabled.',

    'icd10_map_file' :
    'Save converted ICD10 code mappings to file',

    'description_file' :
    'Save descriptions of each column to file',

    'summary_file' :
    'Save summary of cleaning applied to each column to file',

    # TSV export options
    'tsv_sep'    :
    'Column separator string to use in output file (default: "{}")'.format(
        DEFAULT_TSV_SEP.replace('\t', '\\t')),
    'tsv_missing_values' :
    'String to use for missing values in output file (default: empty '
    'string).' ,
    'non_numeric_file' :
    'Export all non-numeric columns (after formatting) to this file instead '
    'of the primary output file.',
    'tsv_var_format' :
    'Apply custom formatter to the specified variable.',

    # HDF5 export options
    'hdf5_key'    :
    'Key/name to use for the HDF5 group (default: '
    '"{}").'.format(DEFAULT_HDF5_KEY),
    'hdf5_style'  :
    'HDF5 style to save as (default: "{}").'.format(DEFAULT_HDF5_STYLE),

    # Miscellaneous options
    'version'      : 'Print version and exit.',
    'dry_run'      : 'Print a summary of what would happen and exit.',
    'no_builtins'  : 'Do not use the built in variable or data coding tables.',
    'low_memory'   : 'Store intermediate results on disk, rather than in RAM. '
                     'Use this flag on systems which cannot store the full '
                     'data set in RAM. ',
    'work_dir'     : 'Directory to store intermediate files (default: '
                     'temporary directory). Only relevant when using '
                     '--low_memory.',
    'log_file'     : 'Save log messages to file.',
    'num_jobs'     : 'Maximum number of jobs to run in parallel. '
                     '(default: {}).'.format(mp.cpu_count()),
    'pass_through' : 'Do not perform any cleaning or processing on the data - '
                     'implies --skip_insertna, --skip_childvalues, '
                     '--skip_clean_funcs, --skip_recoding, and '
                     '--skip_processing.',
    'plugin_file'  : 'Path to file containing custom funpack plugins. Can be '
                     'used multiple times.',
    'overwrite'    : 'Overwrite output file if it already exists',
    'noisy'        : 'Print debug statements. Can be used up to three times.',
    'quiet'        : 'Be quiet.',
}


def makeEpilog():
    """Generates an epilog for the command line help.

    The epilog contains an overview of available plugin functions.
    """

    def genTable(pluginType):
        plugins = custom.listPlugins(pluginType)
        descs   = []

        # The first two lines of all plugin function
        # docstrings are assumed to be:
        #   - function signature
        #   - one line description
        for p in plugins:
            try:
                desc = custom.get(pluginType, p).__doc__.split('\n')
                sig  = desc[0]
                desc = desc[1]
                descs.append((sig, desc))
            except Exception:
                descs.append(('n\a', None))

        maxpluginlen = max([len(p) for p in plugins])

        lines = []
        fmt   = '  - {{:{}s}}  {{}}'.format(maxpluginlen)

        for p, d in zip(plugins, descs):
            sig, desc = d
            lines.append(fmt.format(p, sig))
            if desc is not None:
                lines.append('{}  {}'.format(' ' * maxpluginlen, desc))

        return '\n'.join(lines)

    epilog = 'Available cleaning routines:'   + '\n'   + \
             genTable('cleaner')              + '\n\n' + \
             'Available processing routines:' + '\n'   + \
             genTable('processor')            + '\n'

    return epilog


@ft.lru_cache()
def makeParser(include=None, exclude=None):
    """Creates and returns an ``argparse.ArgumentParser`` for parsing
    ``funpack`` command line arguments.

    :arg include: Configure the parser so it only includes the specified
                  arguments.

    :arg exclude: Configure the parser so it excludes the specified
                  arguments - this overrides ``include``.
    """

    parser = argparse.ArgumentParser(
        'funpack',
        allow_abbrev=False,
        epilog=makeEpilog(),
        formatter_class=argparse.RawDescriptionHelpFormatter)
    helps  = CLI_ARGUMENT_HELP
    descs  = CLI_DESCRIPTIONS

    for group, args in CLI_ARGUMENTS.items():

        desc  = descs.get(group, None)
        group = parser.add_argument_group(group, description=desc)

        for arg, kwargs in args:

            name = arg[-1]

            if exclude is not None and name     in exclude: continue
            if include is not None and name not in include: continue

            if len(arg) == 2:
                arg = ('-{}'.format(arg[0]), '--{}'.format(arg[1]))

            group.add_argument(*arg, help=helps[name], **kwargs)

    return parser


def sanitiseArgs(argv):
    """Sanitises command-line arguments to work around a bug in ``argparse``.

    The ``argparse`` module does not work with non-numeric optional argument
    values that begin with a hyphen, as it thinks that they are argument names
    (see https://bugs.python.org/issue9334).

    This function searches for relevant optional argument names, and prepends
    a space to their values to make sure that ``argparse`` doesn't fall over.

    :arg argv: Command-line arguments
    :returns:  Sanitised command-line arguments
    """
    argv       = list(argv)
    toSanitise = [( '-nv',           [1]),
                  ('--na_values',    [1]),
                  ( '-re',           [1, 2]),
                  ('--recoding',     [1, 2]),
                  ( '-cv',           [2]),
                  ('--child_values', [2])]

    for arg, idxs in toSanitise:

        try:               argidx = argv.index(arg)
        except ValueError: continue

        for i in idxs:

            i   = i + 1 + argidx
            val = argv[i]

            if val.startswith('-'):
                argv[i] = ' ' + val

    return argv


def parseArgsWithConfigFile(argv=None):
    """Checks the command line arguments to see if a configuration file has
    been specified. If so, loads the arguments in the configuration file,
    and then parses the rest of the command line arguments.

    :returns: see :func:`parseArgs`.
    """

    if argv is None:
        argv = sys.argv[1:]

    argv    = sanitiseArgs(argv)
    cfgfile = None

    if '-cfg' in argv or '--config_file' in argv:
        cfgparser = makeParser('config_file')
        cfgfile   = cfgparser.parse_known_args(argv)[0].config_file

    if cfgfile is not None: namespace = loadConfigFile(cfgfile)
    else:                   namespace = None

    return parseArgs(argv, namespace)


def parseArgs(argv=None, namespace=None):
    """Parses ``funpack`` command line arguments.

    :arg argv:      List of arguments to parse.
    :arg namespace: Existing ``argparse.Namespace`` - if not provided, an
                    empty one will be created.
    :returns:       A tuple containing:
                     - an ``argparse.Namespace`` object containing the parsed
                       command-line arguments.
                     - A list of the original arguments that were parsed.
    """

    if argv is None:
        argv = sys.argv[1:]

    argv = sanitiseArgs(argv)
    args = makeParser().parse_args(argv, namespace)

    # error if output file exists
    if (not args.dry_run) and (not args.overwrite) and op.exists(args.outfile):
        print('Output file already exists, and --overwrite was not '
              'specified!')
        sys.exit(1)

    # make input files absolute
    args.infile = [op.realpath(f) for f in args.infile]

    if args.pass_through:
        args.skip_insertna    = True
        args.skip_childvalues = True
        args.skip_clean_funcs = True
        args.skip_recoding    = True
        args.skip_processing  = True

    # the importing.loadData function accepts
    # either a single encoding, or one encoding
    # for each data file.
    if args.encoding is not None:
        if len(args.encoding) == 1:
            args.encoding = args.encoding[0]
        elif len(args.encoding) != len(args.infile):
            raise ValueError('Wrong number of encodings specified - specify '
                             'either one encoding, or one encoding for each '
                             'input file.')

    # parallelisation only allowed
    # in low_memory mode
    if not args.low_memory:
        args.num_jobs = 0

    # turn loaders into dict of { absfile : name } mappings
    if args.loader is not None:
        args.loader = {op.realpath(f) : n for f, n in args.loader}

    # turn index indices into dict of { file : index } mappings
    if args.index is not None:
        args.index = {op.realpath(f) : int(i) for f, i in args.index}

    # turn formatters into dict of { vid : name } mappings
    if args.tsv_var_format is not None:

        tsv_var_format = {}

        for i, (v, n) in enumerate(args.tsv_var_format):

            # Formatters should be set on integer
            # variable IDs. But we allow non-integers
            # to pass through, as the exportData
            # function will also check against column
            # names.
            try:               v = int(v)
            except ValueError: pass

            tsv_var_format[v] = n

        args.tsv_var_format = tsv_var_format

    # turn --subject/--variable/--exclude
    # arguments into lists of IDs. If
    # error is True, an error is raised on
    # unparseable arguments.
    def replaceIDs(things, error=True):
        newthings = []
        failed    = []
        for i, thing in enumerate(things):

            # --subject/--variable/--exclude args
            # may be may be a file name containing
            # a list of IDs
            if op.exists(thing):
                with open(thing, 'rt') as f:
                    parsed = f.read().split()
                    parsed = [int(t.strip()) for t in parsed]

            else:
                # Or they may be an ID or matlab-style
                # start[:step[:stop]] range, both handled
                # by the parseMatlabRange function.
                try:               parsed = util.parseMatlabRange(thing)
                except ValueError: parsed = None

                # Or they may be a comma-separated
                # list of IDs
                if parsed is None:
                    try:
                        parsed = [int(v) for v in thing.split(',')]

                    # --subject may also be an expression,
                    # so if error is False, and the range/
                    # list parses fail, we pass the argument
                    # through. Otherwise we propagate the
                    # error.
                    except ValueError:
                        if error:
                            raise

                if parsed is None:
                    failed.append(thing)
                    continue

            for t in parsed:
                if t not in newthings:
                    newthings.append(t)

        if len(newthings) == 0: newthings = None
        if len(failed)    == 0: failed    = None

        return newthings, failed

    # variable/exclude is transformed into
    # a list of integer IDs, but subject is
    # transformed into a tuple containing
    # ([ID], [exprStr])
    if args.noisy    is     None: args.noisy    = 0
    if args.quiet:                args.noisy    = 0
    if args.subject  is not None: args.subject  = replaceIDs(args.subject,
                                                             False)
    else:                         args.subject  = None, None
    if args.variable is not None: args.variable = replaceIDs(args.variable)[0]
    if args.exclude  is not None: args.exclude  = replaceIDs(args.exclude)[0]

    # The column option accepts
    # column names, or a file
    # containing column names
    def loadIfExists(path):
        if op.exists(path):
            with open(path, 'rt') as f:
                items = f.readlines()
        else:
            items = [path]
        return [i.strip() for i in items]

    if args.column is not None:
        args.column = list(it.chain(*[loadIfExists(c) for c in args.column]))

    # visits are restricted using the
    # keepVisits cleaning function
    if args.visit is not None:
        visit = []
        for v in args.visit:
            if v in ('first', 'last'): visit.append('"{}"'.format(v))
            else:                      visit.append(str(v))

        visit = 'keepVisits({})'.format(','.join(visit))

        if args.global_clean is None: args.global_clean  =       visit
        else:                         args.global_clean += ',' + visit

    # If variables/categories/columns are
    # explicitly specified, remove_unknown
    # is implied.
    if any((args.variable is not None,
            args.category is not None,
            args.column   is not None)):
        args.remove_unknown = True

    # categories can be specified
    # either by name or by ID -
    # convert the latter to integers.
    if args.category is not None:
        for i, c in enumerate(args.category):
            try:               args.category[i] = int(c)
            except ValueError: continue

    # convert rename_column from a sequence of
    # [(oldname, newname)] pairs into a dict of
    # { oldname : newname } mappings.
    if args.rename_column is not None:
        args.rename_column = dict(args.rename_column)

    def numlist(s):
        return np.fromstring(s, sep=',', dtype=np.float)

    # convert na_values from a sequence of [(varid, str)]
    # pairs into a dict of {varid : [value]} mappings
    if args.na_values is not None:
        args.na_values = {int(vid) : numlist(values)
                          for vid, values in args.na_values}

    # Convert recoding from a sequence of [(varid, rawlevels, newlevels)]
    # tuples to a dict of {varid : (rawlevels, newlevels)} mappings
    if args.recoding is not None:
        args.recoding = {int(vid) : (numlist(rawlevels), numlist(newlevels))
                         for vid, rawlevels, newlevels in args.recoding}

    # Convert child_values from a sequence of
    # [(varid, exprs, values)] tuples to a dict of
    # {varid : ([exprs], [values])} mappings
    if args.child_values is not None:
        args.child_values = {int(vid) : (exprs, numlist(values))
                             for vid, exprs, values in args.child_values}

    # convert clean from a sequence of [(varid, expr)]
    # pairs into a dict of {varid : expr} mappings.
    if args.clean is not None:
        args.clean = {int(vid) : expr for vid, expr in args.clean}

    # convert type_clean from a sequence of [(type, expr)]
    # pairs into a dict of {type : expr} mappings.
    if args.type_clean is not None:
        args.type_clean = {util.CTYPES[t] : e for t, e in args.type_clean}

    # The variable_file, datacoding_file, type_file,
    # processing_file, and category_file options
    # can be specified relative to funpackdir
    configdir = op.join(op.dirname(__file__), 'configs')

    def fixPath(f):

        if f is None:
            return f

        options = [
            f,
            op.join(configdir, f),
            op.join(configdir, f.replace('.', op.sep) + '.tsv')
        ]

        for o in options:
            if op.exists(o):
                return o

        # if the fixed version does not
        # exist, allow processing to
        # continue - it will fail later on.
        return f

    if args.variable_file is not None:
        args.variable_file = [fixPath(f) for f in args.variable_file]
    if args.datacoding_file is not None:
        args.datacoding_file = [fixPath(f) for f in args.datacoding_file]

    args.type_file       = fixPath(args.type_file)
    args.processing_file = fixPath(args.processing_file)
    args.category_file   = fixPath(args.category_file)

    return args, argv


def loadConfigFile(cfgfile):
    """Loads arguments from the given configuration file, returning an
    ``argparse.Namespace`` object.
    """

    argv = []

    # If the specified config file does
    # not exist, assume it is a reference
    # to a built-in config file
    if not op.exists(cfgfile):
        if not cfgfile.endswith('.cfg'):
            cfgfile = cfgfile + '.cfg'
        moddir   = op.dirname(op.abspath(__file__))
        cfgfile = op.join(moddir, 'configs', cfgfile)

    log.debug('Loading arguments from configuration file %s', cfgfile)

    with open(cfgfile, 'rt') as f:
        lines = [line.strip() for line in f.readlines()]
        lines = [line         for line in lines if line != '']
        lines = [line         for line in lines if not line.startswith('#')]

    for line in lines:
        words        = list(shlex.split(line))
        name, values = words[0], words[1:]

        argv.append('--{}'.format(name))
        argv.extend(values)

    argv = sanitiseArgs(argv)

    return makeParser(exclude=('outfile', 'infile')).parse_args(argv)



#!/usr/bin/env python
#
# util.py - Miscellaneous utility functions.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains a collection of miscellaneous utility functions and
constants.
"""


import re
import sys
import enum
import time
import logging
import warnings
import functools
import contextlib

try:
    import resource
except ImportError:
    resource = None

import numpy as np


log = logging.getLogger(__name__)


CTYPES = enum.Enum(
    'CTYPES',
    ['sequence',
     'integer',
     'continuous',
     'categorical_single',
     'categorical_single_non_numeric',
     'categorical_multiple',
     'categorical_multiple_non_numeric',
     'time',
     'date',
     'text',
     'compound',
     'unknown'])
"""The ``CTYPES`` enum defines all the types that ``funpack`` is aware of.
"""


DATA_TYPES = {

    # We have to use floating point for
    # integer types because pandas uses
    # nan to represent missing data.
    CTYPES.integer              : np.float32,
    CTYPES.continuous           : np.float32,
    CTYPES.categorical_single   : np.float32,
    CTYPES.categorical_multiple : np.float32,
    CTYPES.sequence             : np.uint32,
}
"""Default internal data type to use for the different variable types.
Used by the :func:`columnTypes` function. These types may be overridden
by the ``InternalType`` column of the variable table, which is populated
from the ``funpack/data/type.txt`` file (see :func:`.loadTableBases`).
"""


def parseColumnName(name):
    """Parses a UK Biobank column name, returns the components.

    Two column naming formats are supported. The name is expected to be either
    a string of the form::

        variable-visit.instance

    Or a string of the form::

        f.variable.visit.instance

    Some variables have the form::

        f.variable..visit.instance

    For these variables, the visit is interpreted as a negative number.

    If ``name`` does not have one of the above forms, a :exc:`ValueError` is
    raised.

    .. note:: For the vast majority of biobank variables, the second number in
              a column name (``visit`` above) corresponds to the assessment
              visit. However, there are a small number of variables which are
              not associated with a specific visit, and thus for which this
              number does not corresopnd to a visit (e.g. variable 40006), but
              to some other coding.

              Confusingly, the UK Biobank showcase refers to the coding that a
              variable adhers to as an "instancing", whilst also using the
              term "instance" to refer to the columns of multi-valued
              variables - the ``instance`` element of the column name.

              The "instancing" that a variable uses is contained in the
              ``Instancing`` column of the variable table.

    :arg name: Column name
    :returns:  A tuple containing:
                - variable ID
                - visit number
                - instance number
    """

    if name.startswith('f'):
        pat = re.compile(r'f\.([0-9]+)\.(\.)?([0-9]+)\.([0-9]+)')
    else:
        pat = re.compile(r'([0-9]+)-(-)?([0-9]+)\.([0-9]+)')

    parts = pat.fullmatch(name)

    if parts is None:
        raise ValueError('Invalid column name: {}'.format(name))

    parts = parts.groups()

    vid      = int(parts[0])
    visit    = int(parts[2])
    instance = int(parts[3])

    if parts[1] is not None:
        visit = -visit

    return (vid, visit, instance)


def generateColumnName(variable, visit, instance):
    """Generate a column name for the given variable, visit and instance.

    :arg variable: Integer variable ID
    :arg visit:    Visit number
    :arg instance: Instance number
    """
    return '{}-{}.{}'.format(variable, visit, instance)


def parseMatlabRange(r):
    """Parses a string containing a MATLAB-style ``start:stop`` or
    ``start:step:stop`` range, where the ``stop`` is inclusive).

    :arg r:   String containing MATLAB_style range.
    :returns: List of integers in the fully expanded range.
    """
    elems = [int(e) for e in r.split(':')]

    if len(elems) == 3:
        start, step, stop = elems
        if   step > 0: stop += 1
        elif step < 0: stop -= 1

    elif len(elems) == 2:
        start, stop  = elems
        stop        += 1
        step         = 1
    elif len(elems) == 1:
        start = elems[0]
        stop  = start + 1
        step  = 1
    else:
        raise ValueError('Invalid range string: {}'.format(r))

    return list(range(start, stop, step))


@contextlib.contextmanager
def timed(op=None, logger=None, lvl=None, fmt=None, minutes=None):
    """Context manager which times a section of code, and prints a log
    message afterwards.

    :arg op:      Name of operation which is being timed

    :arg logger:  Logger object to use - defaults to :attr:`log`.

    :arg lvl:     Log level - defaults to ``logging.INFO``.

    :arg fmt:     Custom message. If not provided, a default message is used.
                  Must be a ``'%'``-style format string which accepts two
                  parameters: the elapsed time (``%s``), and the memory usage
                  (``%i``)..

    :arg minutes: Deprecated, has no effect.
    """

    if minutes is not None:
        warnings.warn('The "minutes" argument is deprecated and has '
                      'no effect', DeprecationWarning, stacklevel=2)
    if fmt is None:
        fmt = '[{}] completed in %s (%+iMB)'.format(op)

    if logger is None:
        logger = log

    if lvl is None:
        lvl = logging.INFO

    if op is not None:
        logger.log(lvl, 'Running task [%s]', op)

    # ru_maxrss appears to be bytes under
    # macos, and kilobytes under linux
    if sys.platform == 'darwin': memdenom = 1048576.0
    else:                        memdenom = 1024.0

    if resource is not None:
        startmem  = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    else:
        startmem = 0

    starttime = time.time()

    yield

    endtime = time.time()

    hours   = int( (endtime - starttime) / 3600)
    minutes = int(((endtime - starttime) % 3600) / 60)
    seconds = int(((endtime - starttime) % 3600) % 60)
    timestr = '{:d} seconds'.format(seconds)

    if minutes > 0: timestr = '{} minutes, {}'.format(minutes, timestr)
    if hours   > 0: timestr = '{} hours, {}'  .format(hours,   timestr)

    if resource is not None:
        endmem  = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    else:
        endmem = 0

    mbytes = (endmem - startmem) / memdenom

    if minutes: logger.log(lvl, fmt, timestr, mbytes)
    else:       logger.log(lvl, fmt, timestr, mbytes)


def deprecated(message):
    """Decorator used to mark a function or method as deprecated """

    def wrapper(func):

        warnings.filterwarnings('default', category=DeprecationWarning)

        def decorator(*args, **kwargs):
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return functools.update_wrapper(decorator, func)

    return wrapper



#!/usr/bin/env python
#
# test_loadtables.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os.path as op
import textwrap
import itertools as it

import pandas as pd
import numpy  as np

import funpack.datatable as datatable
import funpack.loadtables as loadtables
import funpack.util as util
import funpack.custom as custom
import funpack.processing_functions as pfns

from . import tempdir, clear_plugins, patch_base_tables


def _prepare_variableTable(tdir, idxcol=0):

    datadata = [['eid',       '1',    '2',    '3'],
                ['1-0.0',    '10',   '20',   '30'],
                ['2-0.0',   '100',  '200',  '300'],
                ['5-0.0',   '500',  '600',  '700'],
                ['6-0.0',  '1000', '1100', '1200'],
                ['99-0.0',   '99',   '99',   '99']]
    datadata = [['eid',   '1-0.0',  '2-0.0',  '5-0.0',  '6-0.0',  '99-0.0'],
                [  '1',   '10',     '100',    '500',    '1000',   '99'],
                [  '2',   '20',     '200',    '600',    '1100',   '99'],
                [  '3',   '30',     '300',    '700',    '1200',   '99']]

    if idxcol != 0:
        for row in datadata:
            row.insert(idxcol, row.pop(0))

    datadata = '\n'.join(['\t'.join(row) for row in datadata])

    vardata = textwrap.dedent("""
    ID\tType\tDescription\tDataCoding\tNAValues\tRawLevels\tNewLevels\tParentValues\tChildValues\tClean
    1\tInteger\tvar 1\t100\t1,2,3\t10,20\t100,200\tv100 == 20\t123\tkeepVisits('last')
    2\tContinuous\tvar 2\t200\t\t\t\t\t\tremove, keepVisits('first')
    4\tInteger\tvar 4\t\t\t\t\t\t\t
    5\tDate\tvar 5\t\t\t\t\t\t\t
    6\tTime\tvar 6\t\t\t\t\t\t\t
    """).strip()

    dcdata = textwrap.dedent("""
    ID\tNAValues\tRawLevels\tNewLevels
    100\t100,200\t10, 11, 12\t20, 21, 22
    200\t500,600\t20, 21, 22\t30, 31, 32
    """).strip()

    typedata = textwrap.dedent("""
    Type\tClean
    Integer\tmakeNa('< 0')
    Continuous\tmakeNa('== 0')
    """).strip()

    fdata  = op.join(tdir, 'data.tsv')
    fvars  = op.join(tdir, 'vars.tsv')
    fdcs   = op.join(tdir, 'dcs.tsv')
    ftypes = op.join(tdir, 'typs.tsv')

    with open(fdata,  'wt') as f: f.write(datadata)
    with open(fvars,  'wt') as f: f.write(vardata)
    with open(fdcs,   'wt') as f: f.write(dcdata)
    with open(ftypes, 'wt') as f: f.write(typedata)

    return fdata, fvars, fdcs, ftypes


@clear_plugins
def test_loadTables():

    @custom.processor('dummy_proc1')
    def dummy_proc1():
        pass

    @custom.processor('dummy_proc2')
    def dummy_proc2():
        pass

    procdata = textwrap.dedent("""
    Variable\tProcess
    all\tdummy_proc1
    all_independent\tdummy_proc1
    1\tdummy_proc2
    1:5\tdummy_proc1, dummy_proc2
    1,2,3:5\tdummy_proc1
    """).strip()


    catdata = textwrap.dedent("""
    ID\tCategory\tVariables
    1\tcategory 1\t1, 2, 3, 4, 5
    2\tcategory 2\t6, 7, 8, 9, 10
    3\tcategory 3\t11:15, 16, 17, 18, 19, 20
    4\tcategory 4\t21:2:30
    """).strip()

    custom.registerBuiltIns()

    with tempdir() as td, patch_base_tables():

        with open('cattable.tsv', 'wt') as f:
            f.write(catdata)
        with open('proctable.tsv', 'wt') as f:
            f.write(procdata)

        fdata, fvars, fdcs, ftypes = _prepare_variableTable(td)
        vartable, proctable, cattable, unknowns, unprocessed = \
            loadtables.loadTables([fdata],
                                  [fvars],
                                  [fdcs],
                                  ftypes,
                                  'proctable.tsv',
                                  'cattable.tsv')

    assert len(unknowns)   == 1
    assert unknowns[0].vid == 99

    # Test a selection from each table
    expvids = [1, 2, 5, 6, 99]
    expdcs  = [100, 200, None, None, None]
    dcs     = [None if pd.isna(d) else d for d in vartable['DataCoding']]

    assert list(vartable.index) == expvids
    assert dcs                  == expdcs

    expproc4 = ([1, 2, 3, 4, 5],   ['dummy_proc1'])

    vids  = proctable['Variable'][4]
    procs = proctable['Process'][ 4]
    procs = [p.name for p in procs.values()]

    assert vids[0] == 'vids'
    assert vids[1] == expproc4[0]
    assert procs   == expproc4[1]

    assert      cattable['Category'] .iloc[2]  == 'category 3'
    assert list(cattable['Variables'].iloc[2]) == list(range(11, 21))


def test_loadVariableTable_standard():

    custom.registerBuiltIns()

    with tempdir() as td, patch_base_tables():
        fdata, fvars, fdcs, ftypes = _prepare_variableTable(td)
        vartable, unk, unp = loadtables.loadVariableTable(
            [fdata], [fvars], [fdcs], ftypes)

    expvids       = [1, 2, 5, 6, 99]
    exptypes      = [util.CTYPES.integer, util.CTYPES.continuous,
                     util.CTYPES.date, util.CTYPES.time, util.CTYPES.unknown]
    expdescs      = ['var 1', 'var 2', 'var 5', 'var 6', '99-0.0']
    expdcs        = [100, 200, None, None, None]
    expnavals     = [[1, 2, 3], [500, 600], [], [], []]
    exprawlevels  = [[10, 20], [20, 21, 22], [], [], []]
    expnewlevels  = [[100, 200], [30, 31, 32], [], [], []]
    expparentvals = [[100], [], [], [], []]
    expchildvals  = [[123], [], [], [], []]
    expcleans     = [['keepVisits', 'makeNa'],
                     ['remove', 'keepVisits', 'makeNa'], [], [], []]

    assert len(unk)   == 1
    assert unk[0].vid == 99
    assert list(vartable.index)          == expvids
    assert list(vartable['Description']) == expdescs

    types      = [t for t in vartable['Type']]
    dcs        = [None if pd.isna(d) else d for d in vartable['DataCoding']]
    navals     = [list(v) if isinstance(v, np.ndarray) else []
                  for v in vartable['NAValues']]
    rawlevels  = [list(v) if isinstance(v, np.ndarray) else []
                  for v in vartable['RawLevels']]
    newlevels  = [list(v) if isinstance(v, np.ndarray) else []
                  for v in vartable['NewLevels']]
    parentvals = [[]   if pd.isna(exps)
                  else list(it.chain(*[e.variables for e in exps]))
                  for exps in vartable['ParentValues']]
    childvals  = [list(v) if isinstance(v, np.ndarray)
                  else [] for v in vartable['ChildValues']]
    cleans     = [[]   if pd.isna(p) else list(p.keys())
                  for p in vartable['Clean']]

    assert types      == exptypes
    assert dcs        == expdcs
    assert navals     == expnavals
    assert rawlevels  == exprawlevels
    assert newlevels  == expnewlevels
    assert parentvals == expparentvals
    assert childvals  == expchildvals
    assert cleans     == expcleans


def test_loadVariableTable_indexes():


    custom.registerBuiltIns()

    with tempdir() as td:
        fdata, fvars, fdcs, ftypes = _prepare_variableTable(td, 2)
        vartable, unk, unp = loadtables.loadVariableTable(
            [fdata], [fvars], [fdcs], ftypes, indexes={fdata : 2})

        print()
        print(open(fdata).read())

    expvids  = [1, 2, 5, 6, 99]
    expdescs = ['var 1', 'var 2', 'var 5', 'var 6', '99-0.0']

    assert len(unk)   == 1
    assert unk[0].vid == 99
    assert list(vartable.index)          == expvids
    assert list(vartable['Description']) == expdescs


def test_loadVariableTable_clean():

    custom.registerBuiltIns()

    customClean       = {1 : 'fillMissing(10)'}
    customTypeClean   = {util.CTYPES.continuous : 'fillMissing(10)'}
    customGlobalClean = 'convertICD10Codes'

    with tempdir() as td:
        fdata, fvars, fdcs, ftypes = _prepare_variableTable(td)
        ppvt = loadtables.loadVariableTable(
            [fdata], [fvars], [fdcs], ftypes,
            clean=customClean)[0]
        tpvt = loadtables.loadVariableTable(
            [fdata], [fvars], [fdcs], ftypes,
            typeClean=customTypeClean)[0]
        pp_tpvt = loadtables.loadVariableTable(
            [fdata], [fvars], [fdcs], ftypes,
            clean=customClean,
            typeClean=customTypeClean)[0]
        pp_tp_gpvt = loadtables.loadVariableTable(
            [fdata], [fvars], [fdcs], ftypes,
            clean=customClean,
            typeClean=customTypeClean,
            globalClean=customGlobalClean)[0]

    pp_expcleans       = [['fillMissing'],
                          ['remove', 'keepVisits', 'makeNa'],
                          [], [], []]
    tp_expcleans       = [['keepVisits', 'makeNa'],
                           ['remove', 'keepVisits', 'fillMissing'],
                           [], [], []]
    pp_tp_expcleans    = [['fillMissing'],
                           ['remove', 'keepVisits', 'fillMissing'],
                           [], [], []]
    pp_tp_gp_expcleans = [['fillMissing', 'convertICD10Codes'],
                           ['remove', 'keepVisits', 'fillMissing',
                            'convertICD10Codes'],
                           ['convertICD10Codes'],
                           ['convertICD10Codes'],
                           ['convertICD10Codes']]
    pp_cleans          = [[] if pd.isna(p) else list(p.keys())
                           for p in ppvt['Clean']]
    tp_cleans          = [[] if pd.isna(p) else list(p.keys())
                           for p in tpvt['Clean']]
    pp_tp_cleans       = [[] if pd.isna(p) else list(p.keys())
                           for p in pp_tpvt['Clean']]
    pp_tp_gp_cleans    = [[] if pd.isna(p) else list(p.keys())
                           for p in pp_tp_gpvt['Clean']]

    assert pp_cleans       == pp_expcleans
    assert tp_cleans       == tp_expcleans
    assert pp_tp_cleans    == pp_tp_expcleans
    assert pp_tp_gp_cleans == pp_tp_gp_expcleans


@clear_plugins
def test_loadVariableTable_naValues():

    custom.registerBuiltIns()

    with tempdir() as td:
        fdata, fvars, fdcs, ftypes = _prepare_variableTable(td)

        vartable, _, _ = loadtables.loadVariableTable(
            [fdata], [fvars], [fdcs], ftypes,
            naValues={2   : [10, 20, 30],
                      99  : [10, 11, 12],
                      111 : [ 1,  2,  3]})

    expvids   = [1, 2, 5, 6, 99]
    expnavals = [[1, 2, 3], [10, 20, 30], [], [], [10, 11, 12]]

    navals    = [list(v) if isinstance(v, np.ndarray) else []
                 for v in vartable['NAValues']]

    assert list(vartable.index) == expvids
    assert navals               == expnavals


@clear_plugins
def test_loadVariableTable_recoding():

    custom.registerBuiltIns()

    with tempdir() as td:
        fdata, fvars, fdcs, ftypes = _prepare_variableTable(td)

        vartable, _, _ = loadtables.loadVariableTable(
            [fdata], [fvars], [fdcs], ftypes,
            recoding={2   : ([1, 2, 3],    [10, 20, 30]),
                      99  : ([40, 50],     [400, 500]),
                      111 : ([ 1,  2,  3], [100, 200, 300])})

    expvids      = [1, 2, 5, 6, 99]
    exprawlevels = [[10, 20],   [1, 2, 3],    [], [], [40, 50]]
    expnewlevels = [[100, 200], [10, 20, 30], [], [], [400, 500]]

    rawlevels = [list(v) if isinstance(v, np.ndarray) else []
                 for v in vartable['RawLevels']]
    newlevels = [list(v) if isinstance(v, np.ndarray) else []
                 for v in vartable['NewLevels']]

    assert list(vartable.index) == expvids
    assert rawlevels            == exprawlevels
    assert newlevels            == expnewlevels


@clear_plugins
def test_loadVariableTable_childValues():

    custom.registerBuiltIns()

    with tempdir() as td:
        fdata, fvars, fdcs, ftypes = _prepare_variableTable(td)

        childValues = {
            1   : ('v2 != 7', [12345]),
            99  : ('v5  < 7', [54321]),
            111 : ('v1 == 6', [55555])}

        vartable, _, _ = loadtables.loadVariableTable(
            [fdata], [fvars], [fdcs], ftypes,
            childValues=childValues)

    expvids       = [1, 2, 5, 6, 99]
    expparentvals = [[2],     [], [], [], [5]]
    expchildvals  = [[12345], [], [], [], [54321]]

    parentvals = [[]   if pd.isna(exps)
                  else list(it.chain(*[e.variables for e in exps]))
                  for exps in vartable['ParentValues']]
    childvals  = [list(v) if isinstance(v, np.ndarray)
                  else [] for v in vartable['ChildValues']]

    assert list(vartable.index) == expvids
    assert parentvals           == expparentvals
    assert childvals            == expchildvals


@clear_plugins
def test_loadProcessingTable():

    @custom.processor('dummy_proc1')
    def dummy_proc1():
        pass

    @custom.processor('dummy_proc2')
    def dummy_proc2():
        pass

    filedata = textwrap.dedent("""
    Variable\tProcess
    all\tdummy_proc1
    all_independent\tdummy_proc1
    1\tdummy_proc2
    1:5\tdummy_proc1, dummy_proc2
    1,2,3:5\tdummy_proc1
    all_except,1,3\tdummy_proc1
    all_independent_except,2,5\tdummy_proc2
    """).strip()

    exp = [
        ('all',                    [],                ['dummy_proc1']),
        ('all_independent',        [],                ['dummy_proc1']),
        ('vids',                   [1],               ['dummy_proc2']),
        ('vids',                   [1, 2, 3, 4, 5],   ['dummy_proc1',
                                                       'dummy_proc2']),
        ('vids',                   [1, 2, 3, 4, 5],   ['dummy_proc1']),
        ('all_except',             [1, 3],            ['dummy_proc1']),
        ('all_independent_except', [2, 5],            ['dummy_proc2']),
    ]

    prepend = [('100:105', 'dummy_proc1, dummy_proc2')]
    append  = [('100', 'dummy_proc1'), ('200', 'dummy_proc2')]

    preexp = [
        ('vids', [100, 101, 102, 103, 104, 105],
         ['dummy_proc1', 'dummy_proc2'])]
    appexp = [('vids', [100], ['dummy_proc1']),
              ('vids', [200], ['dummy_proc2'])]

    with tempdir():
        with open('proctable.tsv', 'wt') as f:
            f.write(filedata)

        pfns.dummy_proc1 = dummy_proc1
        pfns.dummy_proc2 = dummy_proc2

        pt_normal       = loadtables.loadProcessingTable(
            'proctable.tsv')
        pt_skip         = loadtables.loadProcessingTable(
            'proctable.tsv', skipProcessing=True)
        pt_pre          = loadtables.loadProcessingTable(
            'proctable.tsv', prependProcess=prepend)
        pt_app          = loadtables.loadProcessingTable(
            'proctable.tsv', appendProcess=append)
        pt_pre_app      = loadtables.loadProcessingTable(
            'proctable.tsv',
            prependProcess=prepend,
            appendProcess=append)
        pt_skip_pre_app = loadtables.loadProcessingTable(
            'proctable.tsv',
            prependProcess=prepend,
            appendProcess=append,
            skipProcessing=True)

    assert len(pt_normal)       == len(exp)
    assert len(pt_skip)         == 0
    assert len(pt_pre)          == len(exp) + len(prepend)
    assert len(pt_app)          == len(exp) + len(append)
    assert len(pt_pre_app)      == len(exp) + len(prepend) + len(append)
    assert len(pt_skip_pre_app) ==            len(prepend) + len(append)

    for table, offset in [(pt_normal,  0),
                          (pt_pre,     len(prepend)),
                          (pt_app,     0),
                          (pt_pre_app, len(prepend))]:

        for i in range(len(exp)):

            exppvtype, expvids, expprocs = exp[i]

            pvtype, vids = table['Variable'][i + offset]
            procs        = table['Process'][i + offset]
            procs        = [p.name for p in procs.values()]

            assert pvtype == exppvtype
            assert vids   == expvids
            assert procs  == expprocs

    for table in [pt_pre, pt_pre_app, pt_skip_pre_app]:
        for i in range(len(preexp)):
            exppvtype, expvids, expprocs = preexp[i]

            pvtype, vids = table['Variable'][i]
            procs        = table['Process'][i]
            procs        = [p.name for p in procs.values()]

            assert pvtype == exppvtype
            assert vids   == expvids
            assert procs  == expprocs

    for table, offset in [(pt_app, len(exp)),
                          (pt_pre_app, len(prepend) + len(exp)),
                          (pt_skip_pre_app, len(prepend))]:
        for i in range(len(appexp)):
            exppvtype, expvids, expprocs = appexp[i]

            pvtype, vids = table['Variable'][i + offset]
            procs        = table['Process'][i + offset]
            procs        = [p.name for p in procs.values()]

            assert pvtype == exppvtype
            assert vids   == expvids
            assert procs  == expprocs


def test_loadCategoryTable():

    filedata = textwrap.dedent("""
    ID\tCategory\tVariables
    1\tcategory 1\t1, 2, 3, 4, 5
    2\tcategory 2\t6, 7, 8, 9, 10
    3\tcategory 3\t11:15, 16, 17, 18, 19, 20
    4\tcategory 4\t21:2:30
    """).strip()

    with tempdir():
        with open('cattable1.tsv', 'wt') as f:
            f.write(filedata)

        with open('cattable2.tsv', 'wt') as f:
            f.write(filedata + '\n')
            f.write('5\tunknown\t100,101,102\n')

        unknowns = [datatable.Column('data.tsv', 'var1', 0, 103, 0, 0),
                    datatable.Column('data.tsv', 'var2', 1, 104, 0, 0),
                    datatable.Column('data.tsv', 'var3', 2, 105, 0, 0)]

        cattable1 = loadtables.loadCategoryTable('cattable1.tsv',
                                                 unknownVars=unknowns)
        cattable2 = loadtables.loadCategoryTable('cattable2.tsv',
                                                 unknownVars=unknowns)

    expnames = ['category 1',
                'category 2',
                'category 3',
                'category 4',
                'unknown']
    expvars = [range(1, 6),
               range(6, 11),
               range(11, 21),
               range(21, 30, 2),
               [103, 104, 105]]

    assert len(cattable1) == len(expvars)
    assert len(cattable2) == len(expvars)

    assert np.all(cattable1.index == [1, 2, 3, 4, loadtables.UNKNOWN_CATEGORY_ID])
    assert np.all(cattable2.index == [1, 2, 3, 4, 5])

    for i in range(len(expnames)):

        en =      expnames[i]
        ev = list(expvars[i])

        assert      cattable1['Category'] .iloc[i]  == en
        assert list(cattable1['Variables'].iloc[i]) == ev
        assert      cattable2['Category'] .iloc[i]  == en

        if en == 'unknown':
            assert list(cattable2['Variables'].iloc[i]) == [100, 101, 102] + ev


def test_categoryVariables():
    filedata = textwrap.dedent("""
    ID\tCategory\tVariables
    1\tage, sex\t1, 2, 3, 4, 5
    2\talcohol, tobacco\t6, 7, 8, 9, 10
    3\tmilk\t11:15, 16, 17, 18, 19, 20
    4\tcigars\t7, 8, 9
    """).strip()

    with tempdir():
        with open('cattable.tsv', 'wt') as f:
            f.write(filedata)

        cattable = loadtables.loadCategoryTable('cattable.tsv')

    tests = [
        (['age, sex'],          range(1, 6)),
        (['age', 'sex'],        range(1, 6)),
        (['alcohol'],           range(6, 11)),
        (['age', 'milk'],       it.chain(range(1, 6), range(11, 21))),
        (['milk', 'age'],       it.chain(range(11, 21), range(1, 6))),
        (['age', 1],            range(1, 6)),
        (['age', 3],            it.chain(range(1, 6), range(11, 21))),
        ([3, 'age'],            it.chain(range(11, 21), range(1, 6))),
        (['tobacco', 'cigars'], range(6, 11)),
        (['cigars', 'tobacco'], [7, 8, 9, 6, 10]),
    ]

    for cats, exp in tests:
        vids = loadtables.categoryVariables(cattable, cats)
        exp  = list(exp)
        assert exp == vids



#!/usr/bin/env python
#
# coding.py - Loading files which contain descriptions for data coding values.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains the :func:`loadCodingFile` function, which can be used
to load descriptions of the values for a given data coding.


The data coding information for some data codings in the UKBiobank is stored
in ``funpack/data/coding/`` - these files are available from the UKBiobank
showcase at https://biobank.ctsu.ox.ac.uk/crystal/schema.cgi.
"""


import os.path as op

import pandas  as pd


def getCodingFilePath(dtable=None, vid=None, coding=None):
    """Return a path to the data coding file for the given ``vid``.

    Pass the path to :func:`loadCodingFile` to load it. The returned path
    is not guaranteed to exist.

    :arg dtable: The :class:`.DataTable`
    :arg vid:    The variable ID
    :return:     A path to the relevant data coding file.
    """

    codingdir = op.join(op.dirname(__file__), 'data', 'coding')

    if (dtable is not None) and (vid is not None):
        coding = dtable.vartable.loc[vid, 'DataCoding']

    elif coding is None:
        raise ValueError('Either a datatable+vid, or a data '
                         'coding, must be specified')

    return op.join(codingdir, 'coding{}.tsv'.format(int(coding)))


def loadCodingFile(fname):
    """Loads the given data coding description file.

    The descriptions are returned in a ``pandas.DataFrame``, with the coding
    values as the index, and a single column called ``meaning``, containing
    the descriptions for each value.

    :arg fname: Path to a data coding file
    :return:    A ``DataFrame`` containing descriptions
    """
    return pd.read_csv(fname, delimiter='\t', index_col=0)



#!/usr/bin/env python
#
# test_processing_core.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import itertools as it

import numpy  as np
import pandas as pd


import funpack.util                      as util
import funpack.processing_functions_core as core


def test_isSparse_minpres():

    size = 100
    actual_present    = [0,  0.01, 0.1, 0.5, 1]
    minpres_threshold = [0,  0.01, 0.1, 0.5, 1]

    for present, threshold in it.product(actual_present, minpres_threshold):

        data     = np.random.random(size)
        expected = present < threshold

        missing = int(round(size - present * size))
        missing = np.random.choice(range(size), missing, replace=False)
        data[missing] = np.nan

        data = pd.Series(data)

        absres  = core.isSparse(
            data, util.CTYPES.continuous, minpres=threshold * size)
        propres = core.isSparse(
            data, util.CTYPES.continuous, minpres=threshold, abspres=False)

        if expected:
            expcause = 'minpres'
            expval   =  size - len(missing)
        else:
            expcause = None
            expval   = None

        assert absres  == (expected, expcause, expval)
        assert propres == (expected, expcause, expval)

    # minpres should be ignored if
    # number of points in data is
    # less than or equal to it
    data = np.random.random(10)
    data[:2] = np.nan

    res = core.isSparse(pd.Series(data),
                         util.CTYPES.continuous,
                         minpres=9)
    assert res == (True, 'minpres', 8)

    res = core.isSparse(pd.Series(data),
                        util.CTYPES.continuous,
                        minpres=10)
    assert res == (True, 'minpres', 8)

    res = core.isSparse(pd.Series(data),
                        util.CTYPES.continuous,
                        minpres=11)
    assert res == (False, None, None)

    res = core.isSparse(pd.Series(data),
                        util.CTYPES.continuous,
                        minpres=100)
    assert res == (False, None, None)


def test_isSparse_minstd():

    actualstds = np.linspace(0, 2, 10)
    minstds    = np.linspace(0, 2, 10)

    size = 500

    for actualstd, minstd in it.product(actualstds, minstds):

        data      = np.random.randn(size) * actualstd
        data      = pd.Series(data)
        actualstd = data.std()
        expected  = actualstd <= minstd

        result = core.isSparse(data, util.CTYPES.continuous,
                               minstd=minstd)
        if expected:
            assert result[:2] == (expected, 'minstd')
            assert np.isclose(result[2], actualstd)


def test_isSparse_maxcat():

    size          = 20
    actualmaxcats = np.arange(1, 21, 2)
    maxcats       = np.arange(1, 21, 2)

    dtypes = [util.CTYPES.integer,
              util.CTYPES.categorical_single,
              util.CTYPES.categorical_multiple]

    # test should only be applied
    # for integer/categoricals
    data = pd.Series(np.arange(size))
    result = core.isSparse(data, util.CTYPES.continuous, maxcat=1)
    assert result == (False, None, None)

    # threshold should be ignored if
    # bigger than size of data set
    data[:] = 123
    result = core.isSparse(data, util.CTYPES.integer, maxcat=25)
    assert result == (False, None, None)

    for actualmaxcat, maxcat in it.product(actualmaxcats, maxcats):

        data = np.arange(size)
        data[:actualmaxcat] = size + 1

        data = pd.Series(data)

        expected = actualmaxcat >= maxcat

        if expected:
            expected = (expected, 'maxcat', actualmaxcat)
        else:
            expected = (expected, None, None)

        for dt in dtypes:

            maxcatprop = maxcat / len(data)

            resultabs  = core.isSparse(data, dt, maxcat=maxcat)
            resultprop = core.isSparse(data, dt, maxcat=maxcatprop,
                                       abscat=False)

            assert resultabs  == expected
            assert resultprop == expected

    # when maxcat is a proportion, test should ignore nans
    data        = np.zeros(100, dtype=np.float)
    data[:40]   = np.nan
    data[40:80] = 1
    data        = pd.Series(data)
    assert core.isSparse(data, None, maxcat=0.5, abscat=False) == \
        (True, 'maxcat', 40)
    assert core.isSparse(data, None, maxcat=0.6, abscat=False) == \
        (True, 'maxcat', 40)
    assert core.isSparse(data, None, maxcat=0.7, abscat=False) == \
        (False, None, None)


def test_isSparse_mincat():

    size          = 100
    actualmincats = np.arange(1, 21, 2)
    mincats       = np.arange(1, 21, 2)

    dtypes = [util.CTYPES.integer,
              util.CTYPES.categorical_single,
              util.CTYPES.categorical_multiple]

    # test should only be applied
    # for integer/categoricals
    data     = np.zeros(size)
    data[:5] = size + 1
    data     = pd.Series(data)
    result   = core.isSparse(data, util.CTYPES.continuous,
                             mincat=2)
    assert result == (False, None, None)

    # threshold should be ignored if
    # bigger than size of data set
    data[:] = size + 1
    result = core.isSparse(data, util.CTYPES.integer, mincat=size + 5)
    assert result == (False, None, None)

    for actualmincat, mincat in it.product(actualmincats, mincats):

        data = np.zeros(size)
        data[:actualmincat] = size + 1

        data = pd.Series(data)

        if actualmincat < mincat:
            expected = (True, 'mincat', actualmincat)
        else:
            expected = (False, None, None)

        for dt in dtypes:

            mincatprop = mincat / len(data)

            resultabs  = core.isSparse(data, dt, mincat=mincat)
            resultprop = core.isSparse(data, dt, mincat=mincatprop,
                                       abscat=False)

            assert resultabs  == expected
            assert resultprop == expected

    # when mincat is a proportion, test should ignore nans
    data        = np.zeros(100, dtype=np.float)
    data[:40]   = np.nan
    data[40:60] = 1
    data        = pd.Series(data)
    assert core.isSparse(data, None, mincat=0.1, abscat=False) == \
        (False, None, None)
    assert core.isSparse(data, None, mincat=0.2, abscat=False) == \
        (False, None, None)
    assert core.isSparse(data, None, mincat=0.3, abscat=False) == \
        (False, None, None)
    assert core.isSparse(data, None, mincat=0.4, abscat=False) == \
        (True, 'mincat', 20)


def test_isSparse_non_numeric():


    data     = list(np.random.randint(1, 10, 100))
    data[50] = 'abcde'
    data[51] = np.nan
    data     = pd.Series(data)
    assert core.isSparse(data, minstd=10) == (False, None, None)

    data = pd.Series(['a', 'a', 'b', 'b', 'b', np.nan, np.nan])
    assert core.isSparse(data, mincat=1) == (False, None, None)
    assert core.isSparse(data, mincat=2) == (False, None, None)
    assert core.isSparse(data, mincat=3) == (True, 'mincat', 2)
    assert core.isSparse(data, maxcat=4) == (False, None, None)
    assert core.isSparse(data, maxcat=3) == (True, 'maxcat', 3)



def test_redundantColumns():

    size    = 50
    series1 = np.sin(np.linspace(0, np.pi * 6, size))
    series2 = series1 + np.random.random(size)
    corr = pd.Series(series1).corr(pd.Series(series2))

    data = pd.DataFrame({0 : pd.Series(series1), 1 : pd.Series(series2)})
    cols = [0, 1]

    assert core.redundantColumns(data, cols, corr * 1.01) == []
    assert core.redundantColumns(data, cols, corr * 0.99) == [1]

    # insert some missing values, making
    # sure there are more missing values
    # in series2, and the missingness will
    # be positively correlated
    s1miss = np.random.choice(
        np.arange(size, dtype=np.int), 10, replace=False)
    s2miss = list(s1miss)
    while len(s2miss) < 13:
        idx = np.random.randint(0, size, 1)
        if idx not in s1miss:
            s2miss.append(idx)
    s2miss = np.array(s2miss, dtype=np.int)

    series1[s1miss] = np.nan
    series2[s2miss] = np.nan

    corr   = pd.Series(series1).corr(pd.Series(series2))
    nacorr = np.corrcoef(np.isnan(series1), np.isnan(series2))[0, 1]

    data = pd.DataFrame({0 : pd.Series(series1), 1 : pd.Series(series2)})

    assert core.redundantColumns(data, cols, corr * 1.01) == []
    assert core.redundantColumns(data, cols, corr * 0.99) == [1]

    # both present and missing values must
    # be above the threshold for the column
    # to be considered redundant
    assert core.redundantColumns(data, cols, corr * 0.99, nacorr * 0.99) == [1]
    assert core.redundantColumns(data, cols, corr * 1.01, nacorr * 0.99) == []
    assert core.redundantColumns(data, cols, corr * 0.99, nacorr * 1.01) == []
    assert core.redundantColumns(data, cols, corr * 1.01, nacorr * 1.01) == []

    # the column with more missing values
    # should be the one flagged as redundant
    data = pd.DataFrame({0 : pd.Series(series2), 1 : pd.Series(series1)})
    assert core.redundantColumns(data, cols, corr * 1.01)                == []
    assert core.redundantColumns(data, cols, corr * 0.99)                == [0]
    assert core.redundantColumns(data, cols, corr * 0.99, nacorr * 0.99) == [0]
    assert core.redundantColumns(data, cols, corr * 1.01, nacorr * 0.99) == []
    assert core.redundantColumns(data, cols, corr * 0.99, nacorr * 1.01) == []
    assert core.redundantColumns(data, cols, corr * 1.01, nacorr * 1.01) == []


def test_binariseCategorical():

    data = np.random.randint(1, 10, (100, 10))
    cols = {str(i + 1) : data[:, i] for i in range(10)}
    df   = pd.DataFrame(cols)

    bindata, uniq = core.binariseCategorical(df)

    assert sorted(uniq) == sorted(np.unique(data))

    for i, v in enumerate(uniq):
        exp = np.any(data == v, axis=1)
        assert np.all(bindata[:, i] == exp)

    data[data == 5] = 6
    data[:10, 0] = 5
    cols = {str(i + 1) : data[:, i] for i in range(10)}
    df   = pd.DataFrame(cols)

    bindata, uniq = core.binariseCategorical(df, minpres=11)

    assert sorted(uniq) == [1, 2, 3, 4, 6, 7, 8, 9]

    for i, v in enumerate(uniq):
        exp = np.any(data == v, axis=1)
        assert np.all(bindata[:, i] == exp)


def test_binariseCategorical_missing():

    data = np.full((50, 5), np.nan)

    for i in range(data.shape[1]):
        namask = np.random.random(data.shape[0]) < 0.1
        data[~namask, i] = np.random.randint(1, 10, (~namask).sum())

    expuniq = list(sorted(np.unique(data[~np.isnan(data)])))

    expdata = np.zeros((data.shape[0], len(expuniq)))

    for i, v in enumerate(expuniq):
        expdata[:, i] = np.any(data == v, axis=1)

    cols = {str(i + 1) : data[:, i] for i in range(data.shape[1])}
    df   = pd.DataFrame(cols)

    gotdata, gotuniq = core.binariseCategorical(df)
    assert np.all(gotuniq == expuniq)
    assert np.all(gotdata == expdata)


def test_expandCompound():

    data = []

    for i in range(20):
        rlen = np.random.randint(1, 20)
        row = np.random.randint(1, 100, rlen)
        data.append(row)

    exp = np.full((20, max(map(len, data))), np.nan)

    for i in range(20):
        exp[i, :len(data[i])] = data[i]

    series = pd.Series(data)

    got = core.expandCompound(series)

    expna = np.isnan(exp)
    gotna = np.isnan(got)

    assert np.all(     expna  ==      gotna)
    assert np.all(exp[~expna] == got[~gotna])



#!/usr/bin/env python
#
# processing_functions.py - Processing functions
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains definitions of processing functions - functions which
may be specifeid in the processing table.


A processing function may perform any sort of processing on one or more
variables. A processing function may add, remove, or manipulate the columns of
the :class:`DataTable`.


All processing functions must accept the following as their first two
positional arguments:


 - The :class:`.DataTable` object, containing references to the data, variable,
   and processing table.
 - A list of integer ID of the variables to process.


Furthermore, all processing functions must return one of the following:

 - ``None``, indicating that no columns are to be added or removed.

 - A ``list`` of :class:`.Column` objects describing the columns that should
   be removed from the data.

 - A ``tuple`` of length 2, containing:

    - A list of ``pandas.Series`` that should be added to the data.

    - A list of variable IDs to use for each new ``Series``. This list must
      have the same length as the list of new ``Series``, but if they are not
      associated with any specific variable, ``None`` may be used.

 - A ``tuple`` of length 3, containing:

    - List of columns to be removed
    - List of ``Series`` to be added
    - List of variable IDs for each new ``Series``.

 - A ``tuple`` of length 4, containing the above, and:

    - List of metadata associated with each of the new ``Series``. This will
      be added to the :class:`.Column` objects that represent each of the new
      ``Series``.

The following processing functions are defined:

 .. autosummary::
   :nosignatures:

   removeIfSparse
   removeIfRedundant
   binariseCategorical
   expandCompound
"""


import itertools as it
import              logging
import              collections

import pandas           as pd
import pandas.api.types as pdtypes

from . import processing_functions_core as core
from . import                              custom


log = logging.getLogger(__name__)


@custom.processor()
def removeIfSparse(dtable,
                   vids,
                   minpres=None,
                   minstd=None,
                   mincat=None,
                   maxcat=None,
                   abspres=True,
                   abscat=True,
                   ignoreType=False):
    """removeIfSparse([minpres], [minstd], [mincat], [maxcat], [abspres], [abscat])
    Removes columns deemed to be sparse.

    Removes columns for the variables in ``vids`` if they are sparse.

    :arg ignoreType: Defaults to ``False``. If ``True``, all specified tests are
                     run regardless of the types of the ``vids``.

    See the :func:`.isSparse` function for details on the other arguments.
    """  # noqa

    remove = []

    for vid in vids:

        if ignoreType: vtype = None
        else:          vtype = dtable.vartable.loc[vid, 'Type']

        for col in dtable.columns(vid):

            log.debug('Checking column %s for sparsity', col.name)

            isSparse, test, val = core.isSparse(dtable[:, col.name],
                                                vtype,
                                                minpres=minpres,
                                                minstd=minstd,
                                                mincat=mincat,
                                                maxcat=maxcat,
                                                abspres=abspres,
                                                abscat=abscat)

            if isSparse:
                log.debug('Dropping sparse column %s (%s: %f)',
                          col.name, test, val)
                remove.append(col)

    return remove


@custom.processor()
def removeIfRedundant(dtable, vids, corrthres, nathres=None):
    """removeIfRedundant(corrthres, [nathres])
    Removes columns deemed to be redundant.

    Removes columns from the variables in ``vids`` if they are redundant.

    Redundancy is determined by calculating the correlation between pairs
    of columns - see the :func:`.isRedundant` function.
    """

    # Ignore non-numeric columns
    columns = list(it.chain(*[dtable.columns(v) for v in vids]))
    columns = [c for c in columns
               if pdtypes.is_numeric_dtype(dtable[:, c.name])]

    log.debug('Checking %u columns for redundancy', len(columns))

    columns  = collections.OrderedDict([(c.name, c) for c in columns])
    toremove = core.redundantColumns(
        dtable[:, :], list(columns.keys()), corrthres, nathres)

    if len(toremove) > 0:
        log.debug('Dropping %u redundant columns: %s ...',
                  len(toremove), toremove[:5])

    return [columns[tr] for tr in toremove]


@custom.processor()
def binariseCategorical(dtable,
                        vids,
                        acrossVisits=False,
                        acrossInstances=True,
                        minpres=None,
                        nameFormat=None,
                        replace=True):
    """binariseCategorical([acrossVisits], [acrossInstances], [minpres], [nameFormat], [replace])
    Replace a categorical column with one binary column per category.

    Binarises categorical variables - replaces their columns with
    one new column for each value, containing ``True`` for subjects
    with that value, and ``False`` otherwise.

    :arg dtable:          The :class:`.DataTable`

    :arg vids:            Sequence of variable IDs to (independently) apply the
                          binarisation to.

    :arg acrossVisits:    If ``True``, the binarisation is applied across
                          visits for each variable.

    :arg acrossInstances: If ``True``, the binarisation is applied across
                          instances for each variable.

    :arg minpres:         Optional threshold - categorical values with less
                          than this many occurrences will not be added as
                          columns.

    :arg nameFormat:      Format string defining how the new columns should
                          be named - see below.

    :arg replace:         If ``True`` (the default), the original columns are
                          flagged for removal.

    The ``nameFormat`` argument controls how the new data columns should be
    named - it must be a format string using named replacement fields
    ``'vid'``, ``'visit'``, ``'instance'``, and ``'value'``. The ``'visit'``
    and ``'instance'`` fields may or may not be necessary, depending on the
    value of the ``acrossVisits`` and ``acrossInstances`` arguments.

    The default value for the ``nameFormat`` string is as follows:

    ================ =================== ======================================
    ``acrossVisits`` ``acrossInstances`` ``nameFormat``
    ================ =================== ======================================
    ``False``        ``False``           ``'{vid}-{visit}.{instance}_{value}'``
    ``False``        ``True``            ``'{vid}-{visit}.{value}'``
    ``True``         ``False``           ``'{vid}-{value}.{instance}'``
    ``True``         ``True``            ``'{vid}-0.{value}'``
    ================ =================== ======================================
    """  # noqa

    defaultNameFormat = {
        (False, False) : '{vid}-{visit}.{instance}_{value}',
        (False, True)  : '{vid}-{visit}.{value}',
        (True,  False) : '{vid}-{value}.{instance}',
        (True,  True)  : '{vid}-0.{value}',
    }

    if nameFormat is None:
        nameFormat = defaultNameFormat[acrossVisits, acrossInstances]

    toremove  = []
    newseries = []
    newvids   = []
    newmeta   = []

    for vid in vids:

        visits    = dtable.visits(   vid)
        instances = dtable.instances(vid)
        colgroups = []

        if not (acrossVisits or acrossInstances):
            for visit, instance in it.product(visits, instances):
                colgroups.append(dtable.columns(vid, visit, instance))
        elif acrossInstances and (not acrossVisits):
            for visit in visits:
                colgroups.append(dtable.columns(vid, visit))
        elif (not acrossInstances) and acrossVisits:
            for instance in instances:
                colgroups.append(dtable.columns(vid, instance=instance))
        else:
            colgroups = [dtable.columns(vid)]

        for columns in colgroups:

            data              = dtable[:, [c.name for c in columns]]
            binarised, values = core.binariseCategorical(data, minpres=minpres)

            toremove.extend(columns)

            for col, val in zip(binarised.T, values):

                # make sure no periods appear
                # in the resulting column name.
                # We're assuming here that all
                # categoricals are integers,
                # which has not been verified.
                try:               val = int(val)
                except ValueError: pass

                fmtargs = {
                    'vid'      : str(int(columns[0].vid)),
                    'visit'    : str(int(columns[0].visit)),
                    'instance' : str(int(columns[0].instance)),
                    'value'    : str(val)
                }

                newvids  .append(vid)
                newmeta  .append(val)
                newseries.append(pd.Series(
                    col,
                    index=dtable.index,
                    name=nameFormat.format(**fmtargs)))

    if replace: return toremove, newseries, newvids, newmeta
    else:       return [],       newseries, newvids, newmeta


@custom.processor()
def expandCompound(dtable, vids, nameFormat=None, replace=True):
    """expandCompound([nameFormat], [replace])
    Expand a compound column into a set of columns, one for each value.

    Expands compound variables into a set of columns, one for each value.
    Rows with different number of values are padded with ``np.nan``.

    :arg dtable:     The :class:`.DataTable`

    :arg vids:       Sequence of variable IDs to (independently) apply the
                     expansion to.

    :arg nameFormat: Format string defining how the new columns should be named
                     - see below.

    :arg replace:    If ``True`` (the default), the original columns are
                     flagged for removal.
    """

    if nameFormat is None:
        nameFormat = '{vid}-{visit}.{instance}_{index}'

    columns   = list(it.chain(*[dtable.columns(v) for v in vids]))
    newseries = []
    newvids   = []

    for column in columns:

        data    = dtable[:, column.name]
        newdata = core.expandCompound(data)

        for i in range(newdata.shape[1]):

            coldata = newdata[:, i]
            name    = nameFormat.format(vid=column.vid,
                                        visit=column.visit,
                                        instance=column.instance,
                                        index=i)

            newvids  .append(column.vid)
            newseries.append(pd.Series(coldata,
                                       index=dtable.index,
                                       name=name))

    if replace: return columns, newseries, newvids
    else:       return          newseries, newvids



#!/usr/bin/env python
#
# __init__.py - funpack package
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


__version__ = '1.4.2'
"""The ``funpack`` versioning scheme roughly follows Semantic Versioning
conventions.
"""


from .custom    import (loader,    # noqa
                        sniffer,
                        formatter,
                        exporter,
                        processor,
                        metaproc,
                        cleaner)
from .datatable import (DataTable, # noqa
                        Column)



#!/usr/bin/env python
#
# test_storage.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import multiprocessing as mp
import functools       as ft
import                    warnings

import numpy  as np
import pandas as pd
import           pytest


import funpack.storage as storage

from . import tempdir, gen_test_data


def test_HDFStoreCollection_attrs():
    with tempdir() as td:

        gen_test_data(50, 100, 'data1.txt', start_subj=1)
        gen_test_data(50, 100, 'data2.txt', start_subj=101)

        df1 = pd.read_csv('data1.txt', delimiter='\t', index_col='eid')
        df2 = pd.read_csv('data2.txt', delimiter='\t', index_col='eid')

        store = storage.HDFStoreCollection(prefix='mypref',
                                           workDir=td,
                                           colsPerFile=10)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            store.append(df1)
            store.append(df2)

        assert        store.prefix      == 'mypref'
        assert        store.colsPerFile == 10
        assert        store.workDir     == td
        assert        store.nrows       == 200
        assert    len(store)            == 200
        assert        store.ncols       == 50
        assert np.all(store.columns     == df1.columns)
        assert np.all(store.index       == np.concatenate((df1.index,
                                                           df2.index)))
        for i, col in enumerate(df1.columns):
            assert store.columnIndex(col) == int(i / 10)

        blocks = store.columnBlocks(df1.columns)

        for i in range(5):
            exp = df1.columns[i * 10:i * 10 + 10]
            assert np.all(blocks[i]          == exp)
            assert np.all(store.columnsIn(i) == exp)


def test_HDFStoreCollection_bad_append():
    with tempdir():

        # columns don't match
        gen_test_data(10, 100, 'data1.txt', start_subj=1)
        gen_test_data(5,  100, 'data2.txt', start_subj=101)

        df1 = pd.read_csv('data1.txt', delimiter='\t', index_col='eid')
        df2 = pd.read_csv('data2.txt', delimiter='\t', index_col='eid')

        store = storage.HDFStoreCollection()

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            store.append(df1)

        with pytest.raises(ValueError):
            store.append(df2)


def test_HDFStoreCollection_read():

    with tempdir():

        gen_test_data(20, 100, 'data1.txt', start_subj=1)
        gen_test_data(20, 100, 'data2.txt', start_subj=101)

        df1 = pd.read_csv('data1.txt', delimiter='\t', index_col='eid')
        df2 = pd.read_csv('data2.txt', delimiter='\t', index_col='eid')
        df  = pd.concat((df1, df2), axis=0)

        store = storage.HDFStoreCollection(colsPerFile=5)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            store.append(df1)
            store.append(df2)

        tests = [
            slice(None),
            slice(20, 40),
            slice(20, 40, 2),
            (slice(None),       '1-0.0'),
            (slice(None),       '2-0.0'),
            (slice(None),       '10-0.0'),
            (slice(None),      ['8-0.0', '9-0.0', '10-0.0']),
            (slice(None),      ['4-0.0', '9-0.0', '16-0.0']),
            (slice(20, 30),     '2-0.0'),
            (slice(20, 30, 2),  '2-0.0'),
            (slice(20, 50),    ['2-0.0', '3-0.0']),
            (slice(20, 50),    ['2-0.0', '3-0.0', '6-0.0', '18-0.0']),
        ]

        for test in tests:
            got = store.loc[test]
            exp = df   .loc[test]

            assert np.all(got == exp)


def test_HDFStoreCollection_write():

    with tempdir():

        gen_test_data(20, 50, 'data1.txt', start_subj=1)
        gen_test_data(20, 50, 'data2.txt', start_subj=51)

        df1   = pd.read_csv('data1.txt', delimiter='\t', index_col='eid')
        df2   = pd.read_csv('data2.txt', delimiter='\t', index_col='eid')

        store = storage.HDFStoreCollection(colsPerFile=5)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            store.append(df1)
            store.append(df2)

        # write one column at a time
        for col in df1.columns:

            newdata = pd.Series(np.random.randint(1, 10, 100),
                                index=np.concatenate((df1.index, df2.index)),
                                name=col)

            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                store.loc[:, col] = newdata

            assert np.all(store.loc[:, col] == newdata)

        # write several columns at a time
        tests = [
            [1, 2],
            [1, 6],
            [1, 7, 13, 17, 18],
        ]

        for test in tests:

            cols = ['{}-0.0'.format(t) for t in test]

            newdata = pd.DataFrame(
                np.random.randint(1, 10, (100, len(cols))),
                index=np.concatenate((df1.index, df2.index)),
                columns=cols)

            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                store.loc[:, cols] = newdata

            assert np.all(store.loc[:, cols] == newdata)


def _parallel_read_job(st, col):
    return st.loc[:, col].mean()


def test_HDFStoreCollection_parallel_read():

    with tempdir(), mp.Pool(10) as pool:

        gen_test_data(20, 100, 'data1.txt', start_subj=1)
        gen_test_data(20, 100, 'data2.txt', start_subj=51)

        df1   = pd.read_csv('data1.txt', delimiter='\t', index_col='eid')
        df2   = pd.read_csv('data2.txt', delimiter='\t', index_col='eid')
        df    = pd.concat((df1, df2), axis=0)
        store = storage.HDFStoreCollection(colsPerFile=5)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            store.append(df1)
            store.append(df2)

        got = pool.map(ft.partial(_parallel_read_job, store), df.columns)
        exp = [df[col].mean() for col in df.columns]

        got = np.array(got)
        exp = np.array(exp)

        assert np.all(got == exp)


def _parallel_write_job(st, col):
    data           = st.loc[:, col]
    st.loc[:, col] = data * 2


def test_HDFStoreCollection_parallel_write():

    with tempdir(), mp.Pool(10) as pool:

        mgr  = mp.Manager()

        gen_test_data(20, 100, 'data1.txt', start_subj=1)
        gen_test_data(20, 100, 'data2.txt', start_subj=51)

        df1   = pd.read_csv('data1.txt', delimiter='\t', index_col='eid')
        df2   = pd.read_csv('data2.txt', delimiter='\t', index_col='eid')
        df    = pd.concat((df1, df2), axis=0)
        store = storage.HDFStoreCollection(colsPerFile=5, mgr=mgr)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            store.append(df1)
            store.append(df2)

        pool.map(ft.partial(_parallel_write_job, store), df.columns)

        got = store.loc[:]
        exp = df * 2

        assert np.all(got == exp)



def test_HDFStoreCollection_dropRows():

    with tempdir(), mp.Pool(10) as pool:

        mgr = mp.Manager()

        gen_test_data(50, 500, 'data1.txt', start_subj=1)
        gen_test_data(50, 500, 'data2.txt', start_subj=501)

        df1    = pd.read_csv('data1.txt', delimiter='\t', index_col='eid')
        df2    = pd.read_csv('data2.txt', delimiter='\t', index_col='eid')
        df     = pd.concat((df1, df2), axis=0)
        store1 = storage.HDFStoreCollection(colsPerFile=5, mgr=mgr)
        store2 = storage.HDFStoreCollection(colsPerFile=5, mgr=mgr)

        mask = np.random.choice(np.arange(1, 101), size=50, replace=False)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            store1.append(df1)
            store1.append(df2)
            store2.append(df1)
            store2.append(df2)
            store1.dropRows(mask)
            store2.dropRows(mask, pool)

        exp = df.loc[mask, :]

        assert np.all(store1.index  == exp.index)
        assert np.all(store1.loc[:] == exp.loc[:, :])
        assert np.all(store2.index  == exp.index)
        assert np.all(store2.loc[:] == exp.loc[:, :])

        # test with a boolean mask
        df   = exp
        mask = np.random.randint(1, 11, len(df.index)) > 5

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            store1.dropRows(mask)
            store2.dropRows(mask, pool)

        exp = df.loc[mask, :]
        assert np.all(store1.index  == exp.index)
        assert np.all(store1.loc[:] == exp.loc[:, :])
        assert np.all(store2.index  == exp.index)
        assert np.all(store2.loc[:] == exp.loc[:, :])


def test_HDFStoreCollection_dropColumns():

    with tempdir(), mp.Pool(10) as pool:

        mgr = mp.Manager()

        gen_test_data(50, 500, 'data1.txt', start_subj=1)
        gen_test_data(50, 500, 'data2.txt', start_subj=501)

        df1    = pd.read_csv('data1.txt', delimiter='\t', index_col='eid')
        df2    = pd.read_csv('data2.txt', delimiter='\t', index_col='eid')
        df     = pd.concat((df1, df2), axis=0)
        store1 = storage.HDFStoreCollection(colsPerFile=5, mgr=mgr)
        store2 = storage.HDFStoreCollection(colsPerFile=5, mgr=mgr)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            store1.append(df1)
            store1.append(df2)
            store2.append(df1)
            store2.append(df2)

        drop = np.random.choice(df.columns, size=20, replace=False)
        keep = [c for c in df.columns if c not in drop]

        store1.dropColumns(drop)
        store2.dropColumns(drop, pool)

        exp = df.loc[:, keep]

        assert np.all(store1.columns  == exp.columns)
        assert np.all(store1.loc[:]   == exp.loc[:, :])
        assert np.all(store2.columns  == exp.columns)
        assert np.all(store2.loc[:]   == exp.loc[:, :])



#!/usr/bin/env python
#
# test_coding.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import textwrap as tw
import os.path  as op

import numpy as np

import funpack.coding as coding

from . import (tempdir, gen_DataTable)


def test_getCodingFilePath():

    dt = gen_DataTable([np.random.randint(1, 10, 10)])
    dt.vartable.loc[1, 'DataCoding'] = 4
    exp = op.join(op.dirname(coding.__file__), 'data', 'coding', 'coding4.tsv')

    assert coding.getCodingFilePath(dt, 1)    == exp
    assert coding.getCodingFilePath(coding=4) == exp


def test_loadCodingFile():
    descs = tw.dedent("""
    coding	meaning
    30	meaning 30
    40	meaning 40
    50	meaning 50
    20	meaning 20
    60	meaning 60
    10	meaning 10
    """).strip()

    with tempdir():
        with open('descs.txt', 'wt') as f:
            f.write(descs)
        d = coding.loadCodingFile('descs.txt')
        assert d.loc[30, 'meaning'] == 'meaning 30'



#!/usr/bin/env python
#
# test_expression.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import io

import pytest

import pyparsing as pp

import pandas as pd

import funpack.expression as expression

from . import gen_DataTableFromDataFrame


_test_data = """
index 10 20 30
1 1 2 3
2 4 5 6
3 7  9
4 10 11 12
5 13  15
""".strip().replace(' ', '\t')

_test_cols = {
    10 : '10',
    20 : '20',
    30 : '30',
}


_expr_tests = [
    (  'v10 == 1',  [1, 0, 0, 0, 0]),
    (  'v10 != 1',  [0, 1, 1, 1, 1]),
    (  'v10 >  7',  [0, 0, 0, 1, 1]),
    (  'v10 >= 7',  [0, 0, 1, 1, 1]),
    (  'v10 <  7',  [1, 1, 0, 0, 0]),
    (  'v10 <= 7',  [1, 1, 1, 0, 0]),
    ('~(v10 == 1)', [0, 1, 1, 1, 1]),
    ('~(v10 != 1)', [1, 0, 0, 0, 0]),

    ('v20 == na', [0, 0, 1, 0, 1]),
    ('v20 != na', [1, 1, 0, 1, 0]),

    (  'v10 >= 7 && v30 < 10',                [0, 0, 1, 0, 0]),
    ('~(v10 >= 7 && v30 < 10)',               [1, 1, 0, 1, 1]),
    ( 'v10 >= 4 &&  v30 < 10  || v20 != na',  [1, 1, 1, 1, 0]),
    ('(v10 >= 4 &&  v30 < 10) || v20 != na',  [1, 1, 1, 1, 0]),
    ( 'v10 >= 4 && (v30 < 10  || v20 != na)', [0, 1, 1, 1, 0]),


    # bad
    ('10 == 1',        'error'),
    ('10 ==',          'error'),
    ('v10',            'error'),
    ('v10 ==',         'error'),
    ('v10 1',          'error'),
    ('v10 == 1 &&',    'error'),
    ('v10 == 1 && 24', 'error'),
    ('abcde',          'error'),
]


def test_Expression():

    def vine(e):
        vs = []
        if 'v10' in e: vs.append(10)
        if 'v20' in e: vs.append(20)
        if 'v30' in e: vs.append(30)
        return vs

    data =  pd.read_csv(io.StringIO(_test_data), sep='\t')
    dt = gen_DataTableFromDataFrame(data)

    for expr, expected in _expr_tests:

        if expected == 'error':
            with pytest.raises(pp.ParseException):
                e = expression.Expression(expr)
            continue
        else:
            e = expression.Expression(expr)

        assert sorted(e.variables) == sorted(vine(expr))

        coldata = {vid : col  for vid, col in _test_cols.items()}
        result  = e.evaluate(dt, coldata)

        assert len(result) == len(expected)
        assert all([bool(r) == bool(e) for r, e in zip(result, expected)])


def test_calculaetExpressionEvaluationOrder():

    def makexprs(exprstrs):
        exprs = []
        for exprstr in exprstrs:
            expr = [expression.Expression(e) for e in exprstr.split(',')]
            if len(expr) == 1:
                expr = expr[0]
            exprs.append(expr)
        return exprs

    vids = [1, 2, 3, 4, 5, 7]
    exprs = [
        'v3 == 0',             # 1 depends on 3
        'v3  > 2',             # 2 depends on 3
        'v4 != na',            # 3 depends on 4
        'v5 != na, v6 == na',  # 4 depends on 5 and 6
        'v6 < 0',              # 5 depends on 6
        'v8 == 34'             # 7 depends on 8
    ]

    expected = [
        (0, [1, 2, 7]),
        (1, [3]),
        (2, [4]),
        (3, [5]),
    ]

    exprs = makexprs(exprs)
    result = expression.calculateExpressionEvaluationOrder(vids, exprs)
    assert result == expected


    vids = [1, 2]
    exprs = ['v2 == 20, v3 > 400',
             'v3 < 200']
    expected = [(0, [1]),
                (1, [2])]
    exprs = makexprs(exprs)
    result = expression.calculateExpressionEvaluationOrder(vids, exprs)
    assert result == expected

    vids = [1, 2]
    exprs = ['v3 > 400, v2 == 20',
             'v3 < 200']
    expected = [(0, [1]),
                (1, [2])]
    exprs = makexprs(exprs)
    result = expression.calculateExpressionEvaluationOrder(vids, exprs)
    assert result == expected


    # error on self-dependency
    with pytest.raises(ValueError):
        vids  = [1, 2]
        exprs = ['v1 == 1', 'v4 == 42']
        expression.calculateExpressionEvaluationOrder(vids, makexprs(exprs))

    # error on circular-dependency
    with pytest.raises(ValueError):
        vids  = [1, 2]
        exprs = ['v2 == 1', 'v1 == 2']
        expression.calculateExpressionEvaluationOrder(vids, makexprs(exprs))


    # length mismatch
    with pytest.raises(ValueError):
        vids  = [1, 2, 3]
        exprs = ['v2 == 1']
        expression.calculateExpressionEvaluationOrder(vids, makexprs(exprs))



#!/usr/bin/env python
#
# test_util.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import pytest

import funpack.util as util


def test_parseMatlabRange():
    tests = [
        ('1',      [1]),
        ('1:10',   list(range(1, 11))),
        ('1:2:10', [1, 3, 5, 7, 9]),
        ('10:-1:1', list(range(10, 0, -1))),
        ('10:-2:0', [10, 8, 6, 4, 2, 0]),
    ]

    for rng, exp in tests:
        assert util.parseMatlabRange(rng) == exp

    with pytest.raises(ValueError): util.parseMatlabRange('1:')
    with pytest.raises(ValueError): util.parseMatlabRange('1:1:')
    with pytest.raises(ValueError): util.parseMatlabRange('1:1:1:')
    with pytest.raises(ValueError): util.parseMatlabRange('1:1:1:1')
    with pytest.raises(ValueError): util.parseMatlabRange('abcde')
    with pytest.raises(ValueError): util.parseMatlabRange('a:b')
    with pytest.raises(ValueError): util.parseMatlabRange('a:b:c')


def test_parseColumnName():

    tests = [

        ('0-0.0',     ( 0,   0,  0)),
        ('1-1.1',     ( 1,   1,  1)),
        ('10-10.10',  (10,  10, 10)),
        ('1--1.1',    ( 1,  -1,  1)),
        ('10--10.10', (10, -10, 10)),
        ('f.1.1.1',     ( 1,   1,  1)),
        ('f.10.10.10',  (10,  10, 10)),
        ('f.1..1.1',    ( 1,  -1,  1)),
        ('f.10..10.10', (10, -10, 10)),
    ]

    for col, exp in tests:
        assert util.parseColumnName(col) == exp

    with pytest.raises(ValueError): util.parseColumnName('eid')
    with pytest.raises(ValueError): util.parseColumnName('f.eid')
    with pytest.raises(ValueError): util.parseColumnName('abc')
    with pytest.raises(ValueError): util.parseColumnName('10')
    with pytest.raises(ValueError): util.parseColumnName('10-')
    with pytest.raises(ValueError): util.parseColumnName('10-1')
    with pytest.raises(ValueError): util.parseColumnName('10-1-2')
    with pytest.raises(ValueError): util.parseColumnName('10-1.2.3')
    with pytest.raises(ValueError): util.parseColumnName('f.a')
    with pytest.raises(ValueError): util.parseColumnName('f.a.b')
    with pytest.raises(ValueError): util.parseColumnName('f.a.b.c')
    with pytest.raises(ValueError): util.parseColumnName('f.10.')
    with pytest.raises(ValueError): util.parseColumnName('f.10.10.')
    with pytest.raises(ValueError): util.parseColumnName('f.10.10.10.')
    with pytest.raises(ValueError): util.parseColumnName('f.10.10.10.10')
    with pytest.raises(ValueError): util.parseColumnName('f..10')
    with pytest.raises(ValueError): util.parseColumnName('f.10.10..10')


def test_generateColumnName():

    tests = [

        (( 0,   0,  0), '0-0.0'),
        (( 1,   1,  1), '1-1.1'),
        ((10,  10, 10), '10-10.10'),
        (( 1,  -1,  1), '1--1.1'),
        ((10, -10, 10), '10--10.10'),
    ]

    for var, exp in tests:
        assert util.generateColumnName(*var) == exp


def test_timed():
    with util.timed('abc'):
        pass
    with util.timed('abc'):
        pass
    with util.timed('abc', fmt='abc %s %s'):
        pass
    with util.timed(fmt='abc %s %s'):
        pass



#!/usr/bin/env python
#
# processing_functions_core.py - Functions used by processing_functions.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains utility algorithms and functions used by the
:mod:`processing_functions` module.

 .. autosummary::
   :nosignatures:

   isSparse
   redundantColumns
   binariseCategorical
   expandCompound
"""


import logging
import collections
import itertools as it

import numpy            as np
import pandas           as pd
import pandas.api.types as pdtypes

import funpack.util as util


log = logging.getLogger(__name__)


def isSparse(data,
             ctype=None,
             minpres=None,
             minstd=None,
             mincat=None,
             maxcat=None,
             abspres=True,
             abscat=True):
    """Returns ``True`` if the given data looks sparse, ``False`` otherwise.

    Used by :func:`removeIfSparse`.

    The check is based on the following criteria:

     - The number/proportion of non-NA values must be greater than
       or equal to ``minpres``.

     - The standard deviation of the data must be greater than ``minstd``.

     - For integer and categorical types, the number/proportion of the largest
       category must be less than ``maxcat``.

     - For integer and categorical types, the number/proportion of the largest
       category must be greater than ``mincat``.

    If any of these criteria are not met, the data is considered to be sparse.

    Each criteria can be disabled by passing in ``None`` for the relevant
    parameter.

    By default, the ``minstd`` test is only performed on numeric columns, and
    the ``mincat``/``maxcat`` tests are only performed on integer/categorical
    columns. This behaviour can be overridden by passing in a ``ctype`` of
    ``None``, in which case all speified tests will be performed.

    :arg data:     ``pandas.Series`` containing the data

    :arg ctype:    The series column type (one of the :attr:`.util.CTYPES`
                   values), or ``None`` to disable type-specific logic.

    :arg minpres:  Minimum number/proportion of values which must be present.

    :arg minstd:   Minimum standard deviation, for numeric/categorical types.

    :arg mincat:   Minimum size/proportion of largest category,
                   for integer/categorical types.

    :arg maxcat:   Maximum size/proportion of largest category,
                   for integer/categorical types.

    :arg abspres:  If ``True`` (the default), ``minpres`` is interpreted as
                   an absolute count. Otherwise ``minpres`` is interpreted
                   as a proportion.

    :arg abscat:   If ``True`` (the default), ``mincat`` and ``maxcat`` are
                   interpreted as absolute counts. Otherwise ``mincat`` and
                   ``maxcat`` are interpreted as proportions

    :returns:      A tuple containing:

                    - ``True`` if the data is sparse, ``False`` otherwise.

                    - If the data is sparse, one of ``'minpres'``,
                      ``'minstd'``, ``mincat``, or ``'maxcat'``, indicating
                      the cause of its sparsity. ``None`` if the data is not
                      sparse.

                    - If the data is sparse, the value of the criteria which
                      caused the data to fail the test.  ``None`` if the data
                      is not sparse.
    """

    presmask = data.notnull()
    present  = data[presmask]
    ntotal   = len(data)
    npresent = len(present)

    def fixabs(val, isabs, limit=ntotal, repl=npresent):

        # Turn proportion into
        # an absolute count
        if not isabs:
            val = val * limit
            if (val % 1) >= 0.5: val = np.ceil(val)
            else:                val = np.floor(val)

        # ignore the threshold if it is
        # bigger than the total data length
        if limit < val: return repl
        else:           return val

    iscategorical = ctype in (None,
                              util.CTYPES.integer,
                              util.CTYPES.categorical_single,
                              util.CTYPES.categorical_single_non_numeric,
                              util.CTYPES.categorical_multiple,
                              util.CTYPES.categorical_multiple_non_numeric)
    isnumeric     = ctype in (None,
                              util.CTYPES.continuous,
                              util.CTYPES.integer,
                              util.CTYPES.categorical_single,
                              util.CTYPES.categorical_multiple) and \
                    pdtypes.is_numeric_dtype(data)

    # not enough values
    if minpres is not None:
        if npresent < fixabs(minpres, abspres):
            return True, 'minpres', npresent

    # stddev is not large enough (for
    # numerical/categorical types
    if isnumeric and minstd is not None:
        std = (present - present.mean()).std()
        if std <= minstd:
            return True, 'minstd', std

    # for categorical types
    if iscategorical and ((maxcat is not None) or (mincat is not None)):

        if maxcat is not None:
            maxcat = fixabs(maxcat, abscat, npresent, npresent + 1)
        if mincat is not None:
            mincat = fixabs(mincat, abscat, npresent)

        # mincat - smallest category is too small
        # maxcat - one category is too dominant
        uniqvals   = pd.unique(present)
        uniqcounts = [sum(present == u) for u in uniqvals]
        nmincat    = min(uniqcounts)
        nmaxcat    = max(uniqcounts)

        if mincat is not None:
            if nmincat < mincat:
                return True, 'mincat', nmincat

        if maxcat is not None:
            if nmaxcat >= maxcat:
                return True, 'maxcat', nmaxcat

    return False, None, None


def redundantColumns(data, columns, corrthres, nathres=None):
    """Identifies redundant columns based on the correlation of present and
    missing data.

    :arg data:      ``pandas.DataFrame`` containing the data

    :arg columns:   Sequence of column names to check.

    :arg corrthres: Correlation threshold - columns with a correlation greater
                    than this are identified as redundant.

    :arg nathres:   Correlation threshold to use for missing values. If
                    provided, columns must have a correlation greater than
                    ``corrthres`` *and* a missing-value correlation greater
                    than ``nathres`` to be identified as redundant.

    :returns:       List of redundant column names.
    """

    if nathres is None:
        nathres = 0

    # If not performing na thresholding, we
    # have to test all pairs of columns
    if nathres == 0:
        colpairs = list(it.combinations(columns, 2))

    # But if we are checking the correlation of
    # missing values, we can improve performance
    # by calculating the missingness correlation
    # using a matrix product, then only testing
    # the normal correlation between pairs of
    # columns that exceed the missingness
    # threshold.
    if nathres > 0:

        namask = np.asarray(pd.isna(data[columns]), dtype=np.float16)

        # normalise each column of the namask
        # array to unit standard deviation, but
        # only for those columns which have a
        # stddev greater than 0 (those that
        # correspond to columns with either all
        # missing values, or no missing values).
        nameans              = namask.mean(axis=0)
        nastds               = namask.std( axis=0)
        nahasdev             = nastds > 0
        namask              -= nameans
        namask[:, nahasdev] /= nastds[nahasdev]
        nacorr               = np.dot(namask.T, namask) / len(data)

        # Only pairs which exceed the missing
        # threshold need to be further tested.
        # And as the matrix is symmetric, we
        # can drop column pairs where x <= y.
        colpairs = zip(*np.where(nacorr > nathres))
        colpairs = [(x, y) for x, y in colpairs if x > y]
        colpairs = [(columns[x], columns[y]) for x, y in colpairs]

    redundant = set()

    # calculate correlation between column pairs
    for coli, colj in colpairs:

        if coli in redundant or colj in redundant:
            continue

        datai = data[coli]
        dataj = data[colj]
        corr  = datai.corr(dataj)

        # i and j are highly correlated -
        # flag the one with more missing
        # values as redundant
        if corr > corrthres:
            if sum(datai.isna()) > sum(dataj.isna()): drop, keep = coli, colj
            else:                                     drop, keep = colj, coli

            log.debug('Column %s is redundant (correlation with %s: %f)',
                      drop, keep, corr)
            redundant.add(drop)

    return list(redundant)


def binariseCategorical(data, minpres=None):
    """Takes one or more columns containing categorical data,, and generates a
    newd set of binary columns, one for each unique categorical value,
    containing ``True`` in rows where the value was present, ``False``
    otherwise.

    :arg data:    A ``pandas.DataFrame``containing the input columns
    :arg minpres: Optional threshold - values with less than this number of
                  occurrences will not be included in the output

    :returns:     A tuple containing:

                   - A ``(nrows, nvalues)`` ``numpy`` array containing the
                     generated binary columns

                   - A 1D ``numpy`` array of length ``nvalues`` containing
                     the unique values that are encoded in the binary columns.
    """

    if minpres is None:
        minpres = 0

    cols    = [data[c]      for c in data.columns]
    cols    = [c[c.notna()] for c in cols]
    coluniq = [np.unique(c, return_counts=True) for c in cols]
    uniq    = collections.defaultdict(lambda : 0)

    for coluniq, colcounts in coluniq:
        for val, count in zip(coluniq, colcounts):
            uniq[val] += count

    # drop values with low occurrence
    uniq = [v for v, c in uniq.items() if c >= minpres]
    uniq = list(sorted(uniq))

    # Make a new subjects * values array,
    # where each column is True where the
    # subject had an entry equal to that
    # value, False otherwise.
    bindata = np.zeros((len(data), len(uniq)), dtype=np.uint8)
    for i, v in enumerate(uniq):
        bindata[:, i] = (data == v).any(axis=1)

    return bindata, uniq


def expandCompound(data):
    """Takes a ``pandas.Series`` containing sequence data (potentially of
    variable length), and returns a 2D ``numpy`` array containing the parsed
    data.

    The returned array has shape ``(X, Y)``, where ``X`` is the number of rows
    in the data, and ``Y`` is the maximum number of values for any of the
    entries.

    :arg data: ``pandas.Series`` containing the compound data.
    :returns:   2D ``numpy`` array containing expanded data.
    """

    nrows = len(data)
    lens  = data.apply(len).values
    ncols = max(lens)

    # Create a 2D array from
    # rows of different lengths
    #
    # https://stackoverflow.com/a/32043366
    #
    # 2D mask array of shape (nrows, max(lens))
    # which identifies positions to be filled
    mask          = np.arange(ncols) < np.atleast_2d(lens).T
    newdata       = np.full((nrows, ncols), np.nan, dtype=np.float32)
    newdata[mask] = np.concatenate(data.values)

    return newdata



#!/usr/bin/env python
#
# test_config.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os.path as op

import funpack.config as config
import funpack.custom as custom

from . import clear_plugins

@clear_plugins
def test_parseArgs_fixPath():

    custom.registerBuiltIns()

    fullpath = op.normpath(
        op.join(op.dirname(__file__), '..',
                'configs', 'fmrib', 'variables_clean.tsv'))
    relpath  = op.join('fmrib', 'variables_clean.tsv')
    dotpath  = 'fmrib.variables_clean'

    argv = ['-vf', fullpath,
            '-vf', relpath,
            '-vf', dotpath, 'output',  'input']

    args = config.parseArgs(argv)[0]

    assert args.variable_file == [fullpath, fullpath, fullpath]



#!/usr/bin/env python
#
# test_fileinfo.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import textwrap

import pytest

import funpack.custom as custom
import funpack.fileinfo as fileinfo
import funpack.datatable as datatable

from . import tempdir, clear_plugins


AVID = datatable.AUTO_VARIABLE_ID


sniff_tests = [
    ("""col1, col2, col3
1, 2, 3
4, 5, 6
""",
     [('col1', None, None, None),
      ('col2', None, None, None),
      ('col3', None, None, None)]),
    ("""1, 2, 3
4, 5, 6
7, 8, 9
""",
     [(None, None, None, None),
      (None, None, None, None),
      (None, None, None, None)]),
    ("""col1\tcol2\tcol3
1\t2\t3
4\t5\t6
""",
     [('col1', None, None, None),
      ('col2', None, None, None),
      ('col3', None, None, None)]),
    ("""eid\t1-0.0\t2-1.2\t3-3.5
1\t2\t3\t4
5\t6\t7\t8
9\t10\t11\t12
""",
     [('eid',   None, None, None),
      ('1-0.0', 1, 0, 0),
      ('2-1.2', 2, 1, 2),
      ('3-3.5', 3, 3, 5)]),
    ("""1000   2000   3000
4000   5000   6000
7000   8000   9000
""",
     [(None, None, None, None),
      (None, None, None, None),
      (None, None, None, None)]),
    ("""col1   col2   col3
1000   2000   3000
4000   5000   6000
7000   8000   9000
""",
     [('col1', None, None, None),
      ('col2', None, None, None),
      ('col3', None, None, None)]),

    ("""col1  col2      col3
1000   2000   3000
4000   5000   6000
7000   8000   9000
""",
     [('col1', None, None, None),
      ('col2', None, None, None),
      ('col3', None, None, None)]),

    ("""col1\tcol2\tcol3
\t\t
1\t\t
1\t2\t3
1\t2\t
\t2\t3
\t\t3
""",
     [('col1', None, None, None),
      ('col2', None, None, None),
      ('col3', None, None, None)]),
    ("""1
2
3
4
5
6
""",
     [(None, None, None, None)]),
    ("""col1
2
3
4
5
6
""",
     [('col1', None, None, None)]),
    ("""1-2.3
2
3
4
5
6
""",
     [('1-2.3', 1, 2, 3)]),
    ("""eid
2
3
4
5
6
""",
     [('eid', None, None, None)]),
    ("""eid





""",
     [('eid', None, None, None)]),
    ("""eid\t1-0.0\t2-0.0
\t\t
\t\t
\t\t
\t\t
\t\t
""",
     [('eid',   None, None, None),
      ('1-0.0', 1, 0, 0),
      ('2-0.0', 2, 0, 0)]),
    ("""eid,1-0.0,2-0.0
1,,
2,,
3,,
4,,
5,,3
""",
     [('eid',   None, None, None),
      ('1-0.0', 1, 0, 0),
      ('2-0.0', 2, 0, 0)]),
]


def test_sniff():

    for test, expected in sniff_tests:

        with tempdir():

            with open('data.txt', 'wt') as f:
                f.write(test)

            dialect, cols = fileinfo.sniff('data.txt')

        assert len(cols) == len(expected)

        for i, (exp, col) in enumerate(zip(expected, cols)):

            name, vid, visit, instance = exp

            assert col.datafile == 'data.txt'
            assert col.index    == i
            assert col.name     == name
            assert col.vid      == vid
            assert col.visit    == visit
            assert col.instance == instance


    baddata = textwrap.dedent("""
    abcdefg
    1, 2, 3
    4, 5\t8
    """).strip()


    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(baddata)

        with pytest.raises(ValueError):
            fileinfo.sniff('data.txt')



fileinfo_tests = [
    ("""col1, col2, col3
1, 2, 3
4, 5, 6
""",
     True,
     [('col1', 0,        0, 0),
      ('col2', AVID,     0, 0),
      ('col3', AVID + 1, 0, 0)]),
    ("""1, 2, 3
4, 5, 6
7, 8, 9
""",
     False,
     [( '0-0.0' .format(0),        0,        0, 0),
      ( '{}-0.0'.format(AVID + 2), AVID + 2, 0, 0),
      ( '{}-0.0'.format(AVID + 3), AVID + 3, 0, 0)]),
    ("""col1\tcol2\tcol3\tcol4
1\t2\t3\t4
5\t6\t7\t8
""",
     True,
     [('col1', 0,        0, 0),
      ('col2', AVID + 4, 0, 0),
      ('col3', AVID + 5, 0, 0),
      ('col4', AVID + 6, 0, 0)]),
    ("""eid\t1-0.0\t2-1.2\t3-3.5
1\t2\t3\t4
5\t6\t7\t8
9\t10\t11\t12
""",
     True,
     [('eid',   0, 0, 0),
      ('1-0.0', 1, 0, 0),
      ('2-1.2', 2, 1, 2),
      ('3-3.5', 3, 3, 5)]),
    ("""1
2
3
4
5
6
""",
     False,
     [('0-0.0', 0, 0, 0)]),
    ("""col1
2
3
4
5
6
""",
     True,
     [('col1', 0, 0, 0)]),
    ("""eid
2
3
4
5
6
""",
     True,
     [('eid', 0, 0, 0)]),
]


def test_fileinfo():

    with tempdir():

        fnames = []

        for i, test in enumerate(fileinfo_tests):
            fnames.append('data{}.txt'.format(i))
            with open(fnames[-1], 'wt') as f:
                f.write(test[0])

        dialects, headers, colss = fileinfo.fileinfo(fnames)

        dialects = list(dialects)
        headers  = list(headers)
        colss    = list(colss)

        fnames.append(fnames[-1])
        sd, sh, sc = fileinfo.fileinfo(fnames[-1])
        dialects.append(sd)
        headers .append(sh)
        colss   .append(sc)

    assert len(dialects) == len(fnames)
    assert len(headers)  == len(fnames)
    assert len(colss)    == len(fnames)

    for ti, test in enumerate(fileinfo_tests):

        exphdr  = test[1]
        expcols = test[2]

        assert headers[ti]    == exphdr
        assert len(colss[ti]) == len(expcols)

        for i, (exp, col) in enumerate(zip(expcols, colss[ti])):

            name, vid, visit, instance = exp

            assert col.datafile == 'data{}.txt'.format(ti)
            assert col.index    == i
            assert col.name     == name
            assert col.vid      == vid
            assert col.visit    == visit
            assert col.instance == instance


def test_fileinfo_indexes():
    data = textwrap.dedent("""
    col1,col2
    1,4
    2,5
    3,6
    """.strip())
    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write(data)

        cols = fileinfo.fileinfo('data.txt')[2][0]
        assert cols[0].vid == 0
        assert cols[1].vid == AVID

        cols = fileinfo.fileinfo('data.txt', indexes={'data.txt' : 0})[2][0]
        assert cols[0].vid == 0
        assert cols[1].vid == AVID

        cols = fileinfo.fileinfo('data.txt', indexes={'data.txt' : 1})[2][0]
        assert cols[0].vid == AVID
        assert cols[1].vid == 0



@clear_plugins
def test_fileinfo_parser():

    customCols = [
        datatable.Column('custom.txt', 'col1', 0, 10, 0, 0),
        datatable.Column('custom.txt', 'col2', 1, 11, 0, 0),
        datatable.Column('custom.txt', 'col3', 2, 12, 0, 0)]

    @custom.sniffer('test_fileinfo')
    def columns(f):
        return customCols


    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(fileinfo_tests[0][0])

        fnames  = ['data.txt', 'custom.txt']
        parsers = {'custom.txt' : 'test_fileinfo'}

        dialects, headers, colss = fileinfo.fileinfo(
            fnames, sniffers=parsers)

    assert len(dialects) == 2
    assert len(headers)  == 2
    assert len(colss)    == 2

    expcols = fileinfo_tests[0][2]

    for i, (exp, col) in enumerate(zip(expcols, colss[0])):

        name, vid, visit, instance = exp

        assert col.datafile == 'data.txt'
        assert col.index    == i
        assert col.name     == name
        assert col.vid      == vid
        assert col.visit    == visit
        assert col.instance == instance

    assert colss[1] == customCols



#!/usr/bin/env python
#
# processing.py - Cleaning and processing parsing and functions.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functionality for parsing the ``Process`` column of
the processing table, and the ``Clean`` column of the variable
table. Definitions of the available (pre-)processing functions are in the
:mod:`cleaning_functions` and :mod:`.processing_functions` modules.


The :func:`processData` function is also defined here - it executes the
processes defined in the processing table.


Special processing functions can be applied to a variable's data by adding
them to the ``Clean`` and ``Process`` columns of the variable or processing
table respectively.  Processing is specified as a comma-separated list of
process functions - for example::


    process1, process2(), process3('arg1', arg2=1234)


The :func:`parseProcesses` function parses such a line, and returns a list of
:class:`Process` objects which can be used to query the process name and
arguments, and to run each process.
"""


import functools as ft
import itertools as it
import              logging
import              collections

import pyparsing as pp

from . import util
from . import custom


log = logging.getLogger(__name__)


def processData(dtable):
    """Applies all processing specified in the processing table to the data.

    :arg dtable: The :class:`DataTable` instance.
    """

    ptable = dtable.proctable

    for i in ptable.index:

        # refresh the list of all variables
        # in the data on each iteration, as
        # a previously executed process may
        # add/remove variables to/from the
        # data.
        all_vids     = dtable.variables
        all_vids     = [v for v in all_vids if v != 0]

        # For each process, the processing table
        # contains a "process variable type",
        # a list of vids, and the process itself.
        # The pvtype is one of:
        #   - vids:                   apply the process to the specified vids
        #   - all:                    apply the process to all vids
        #   - all_independent:        apply the process independently to all
        #                             vids
        #   - all_except:             apply the process to all vids except the
        #                             specified ones
        #   - all_independent_except: apply the process independently to all
        #                             vids except the specified ones
        pvtype, vids = ptable.loc[i, 'Variable']
        procs        = ptable.loc[i, 'Process']

        # Build a list of lists of vids, with
        # each vid list a group of variables
        # that is to be processed together.

        # apply independently to every variable
        if pvtype in ('all_independent', 'all_independent_except'):
            if pvtype.endswith('except'): exclude = vids
            else:                         exclude = []
            vids = [[v] for v in all_vids if v not in exclude]

        # apply simultaneously to every variable
        elif pvtype in ('all', 'all_except'):
            if pvtype.endswith('except'): exclude = vids
            else:                         exclude = []
            vids = [[v for v in all_vids if v not in exclude]]

        # apply to specified variables
        else:
            vids = [[v for v in vids if dtable.present(v)]]

        vids = [vg for vg in vids if len(vg) > 0]

        if len(vids) == 0:
            continue

        # Run each process
        for proc in procs.values():
            runProcess(proc, dtable, vids)


def runProcess(proc, dtable, vids):
    """Called by :func:`processData`. Runs the given process, and uppates
    the :class:`.DataTable` as needed.

    :arg proc:   :class:`.Process` to run.
    :arg dtable: :class:`.DataTable` containing the data.
    :arg vids:   List of lists, groups of variable IDs to run the process on.
    """

    # We assume that processes which work on more
    # than one variable will manage their own
    # concurrency. Processes which work on just
    # one variable are executed in parallel here.
    runparallel = [vg for vg in vids if len(vg) == 1]
    runserial   = [vg for vg in vids if len(vg)  > 1]

    log.debug('Running process %s on %u variables',
              proc.name, len(list(it.chain(*vids))))

    with util.timed(proc.name, log, logging.DEBUG):

        with dtable.pool() as pool:
            results = list(pool.map(ft.partial(proc.run, dtable), runparallel))

        for vg in runserial:
            results.append(proc.run(dtable, vg))

    remove  = []
    add     = []
    addvids = []
    addmeta = []

    def expand(res, length):
        if res is None: return [None] * length
        else:           return res

    for result in results:
        if result is None:
            continue

        error = ValueError('Invalid return value from '
                           'process {}'.format(proc.name))

        if isinstance(result, tuple):

            # series/vids to add
            if len(result) == 2:
                add    .extend(result[0])
                addvids.extend(expand(result[1], len(result[0])))
                addmeta.extend(expand(None,      len(result[0])))

            # columns to remove, and
            # series/vids to add
            elif len(result) in (3, 4):

                if len(result) == 3:
                    result = list(result) + [None]

                remove .extend(result[0])
                add    .extend(result[1])
                addvids.extend(expand(result[2], len(result[1])))
                addmeta.extend(expand(result[3], len(result[1])))
            else:
                raise error

        # columns to remove
        elif isinstance(result, list):
            remove.extend(result)
        else:
            raise error

    # remove columns first, in case
    # there is a name clash between
    # the old and new columns.
    if len(remove) > 0: dtable.removeColumns(remove)
    if len(add)    > 0: dtable.addColumns(add, addvids, addmeta)


class NoSuchProcessError(Exception):
    """Exception raised by the :class:`Process` class when an unknown
    process name is specified.
    """
    pass


class Process(object):
    """Simple class which represents a single processing step. The :meth:`run`
    method can be used to run the process on the data for a specific variable.
    """


    def __init__(self, ptype, name, args, kwargs):
        """Create a ``Process``.

        :arg ptype: Process type - either ``cleaner`` or ``processor``
                    (see the :mod:`.custom` module).
        :arg name:  Process name
        :arg args:  Positional arguments to pass to the process function.
        :arg args:  Keyword arguments to pass to the process function.
        """

        # cleaner functions are not
        # defined in processing_functions,
        # so in this case func will be None.
        self.__ptype    = ptype
        self.__name     = name
        self.__args     = args
        self.__kwargs   = kwargs
        self.__metaproc = kwargs.pop('metaproc', None)


    def __repr__(self):
        """Return a string representation of this ``Process``."""
        args    = ','.join([str(v) for v in self.__args])
        kwargs  = ','.join(['{}={}'.format(k, v) for k, v in
                           self.__kwargs.items()])

        allargs = [args, kwargs]
        allargs = [a for a in allargs if a != '']
        allargs = ', '.join(allargs)
        return '{}[{}]({})'.format(self.__name, self.__ptype, allargs)


    @property
    def name(self):
        """Returns the name of this ``Process``. """
        return self.__name


    @property
    def args(self):
        """Returns the positional arguments for this ``Process``. """
        return self.__args


    @property
    def kwargs(self):
        """Returns the keyword arguments for this ``Process``. """
        return self.__kwargs


    def run(self, *args):
        """Run the process on the data, passing it the given arguments,
        and any arguments that were passed to :meth:`__init__`.
        """
        result = custom.run(self.__ptype,
                            self.__name,
                            *args,
                            *self.__args,
                            **self.__kwargs)

        if self.__metaproc is not None and \
           isinstance(result, tuple)   and \
           len(result) == 4:

            # The first argument to a process
            # should be the data table
            dtable = args[0]

            # The 3rd/4th args returned from a
            # process hould be a list of vids,
            # and a list of metadata for each of
            # them
            vids    = result[2]
            meta    = result[3]
            mproc   = self.__metaproc
            newmeta = []

            for vid, vmeta in zip(vids, meta):
                try:
                    newval = custom.runMetaproc(mproc, dtable, vid, vmeta)
                    newmeta.append(newval)

                except Exception as e:
                    log.warning('Metadata processing function failed: %s', e)
                    newmeta.append(vmeta)

            result = tuple(list(result[:3]) + [newmeta])

        return result


def parseProcesses(procs, ptype):
    """Parses the given string containing one or more comma-separated process
    calls, as defined in the processing table. Returns a list of
    :class:`Process` objects.

    :arg procs: String containing one or more comma-separated (pre-)processing
                steps.

    :arg ptype: either ``cleaner`` or ``processor``

    :returns:   A list of :class:`Process` objects.

    """

    def makeProcess(toks):
        name = toks[0]

        args   = ()
        kwargs = {}

        if len(toks) == 2:
            if isinstance(toks[1], tuple):
                args   = toks[1]
            elif isinstance(toks[1], dict):
                kwargs = toks[1]
        elif len(toks) == 3:
            args, kwargs = toks[1:]

        if not custom.exists(ptype, name):
            raise NoSuchProcessError(name)

        return Process(ptype, name, args, kwargs)

    parser = pp.delimitedList(makeParser().setParseAction(makeProcess))

    try:
        parsed = parser.parseString(procs, parseAll=True)
    except Exception as e:
        log.error('Error parsing process list "{}": {}'.format(procs, e))
        raise e

    return list(parsed)


@ft.lru_cache()
def makeParser():
    """Generate a ``pyparsing`` parser which can be used to parse a single
    process call in the processing table.
    """

    lparen   = pp.Literal('(').suppress()
    rparen   = pp.Literal(')').suppress()

    def convertBoolean(tok):
        tok = tok[0]
        if   tok == 'True':  return True
        elif tok == 'False': return False
        else:                return tok

    def parseArgs(toks):
        return [tuple(toks)]

    def parseKwargs(toks):
        kwargs = collections.OrderedDict()
        for i in range(0, len(toks), 2):
            kwargs[toks[i]] = toks[i + 1]
        return kwargs

    funcName = pp.pyparsing_common.identifier
    argval   = (pp.QuotedString('"')                                       ^
                pp.QuotedString("'")                                       ^
                pp.pyparsing_common.number                                 ^
                pp.oneOf(['True', 'False']).setParseAction(convertBoolean) ^
                pp.Literal('None').setParseAction(pp.replaceWith(None)))
    kwargs   = (pp.pyparsing_common.identifier +
                pp.Literal('=').suppress() +
                argval)
    posargs  = pp.delimitedList(argval).setParseAction(parseArgs)
    kwargs   = pp.delimitedList(kwargs).setParseAction(parseKwargs)
    allargs  = pp.delimitedList(pp.Optional(posargs) + pp.Optional(kwargs))
    allargs  = lparen   + pp.Optional(allargs) + rparen
    function = funcName + pp.Optional(allargs)

    return function



#!/usr/bin/env python
#
# fmrib.py - Custom loaders for FMRIB-speciifc files.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains custom loaders for data files used in FMRIB,
which contain data on imaged subjects.
"""


import functools as ft
import datetime  as dt
import              calendar

import pandas    as pd
import numpy     as np

import funpack


@funpack.sniffer('FMRIBImaging')
@ft.lru_cache()
def columns_FMRIBImaging(infile):
    """Return a list of columns describing the contents of the
    ``FMRIB_internal_info.txt`` file.
    """

    names = ['eid',
             'acq_time',
             'acq_phase',
             'processing_phase',
             'flipped_SWI']

    return [funpack.Column(infile,
                           n,
                           i,
                           200000 + i,
                           0,
                           0) for i, n in enumerate(names)]


@funpack.loader('FMRIBImaging')
def load_FMRIBImaging(infile):
    """Load a file with the same format as the ``FMRIB_internal_info.txt``
    file.
    """

    def parse_acq_date(date):
        year  = int(date[ :4])
        month = int(date[4:6])
        day   = int(date[6:8])
        return dt.date(year, month, day)

    def parse_acq_time(time):
        hour       = int(time[ :2])
        minute     = int(time[2:4])
        second     = int(time[4:6])
        micro      = int(time[7:])
        return dt.time(hour, minute, second, micro,
                       dt.timezone(dt.timedelta(0)))

    def combine_datetime(date, time):
        date = date.to_pydatetime().date()
        time = time.to_pydatetime().timetz()
        return dt.datetime.combine(date, time)

    names = ['eid',
             'acq_date',
             'acq_time',
             'acq_phase',
             'processing_phase',
             'flipped_SWI']

    converters = {'acq_date' : parse_acq_date,
                  'acq_time' : parse_acq_time}

    df = pd.read_csv(infile,
                     header=None,
                     names=names,
                     index_col=0,
                     parse_dates=['acq_time', 'acq_date'],
                     converters=converters,
                     delim_whitespace=True)

    df['acq_time'] = df['acq_date'].combine(df['acq_time'], combine_datetime)
    df.drop('acq_date', axis=1, inplace=True)

    return df


def dateAsYearFraction(d):
    """Normalise dates so they are represented as a year plus year fraction.
    """
    try:
        # this normalisation results in
        # a non-linear representation of
        # time, but this is apparently
        # not important.
        d          = d.timetuple()
        daysinyear = 366 if calendar.isleap(d.tm_year) else 365
        return d.tm_year + (d.tm_yday - 1) / float(daysinyear)
    except Exception:
        return np.nan


@funpack.formatter('FMRIBImagingDate')
def format_dateAsYearFraction(dtable, column, series):
    """Formats dates using :func:`dateAsYearFraction`. """
    return series.apply(dateAsYearFraction)


def normalisedAcquisitionTime(t):
    """Normalises timestamps so they are represented as a year plus year
    fraction, where days are normalised to lie between 7am and 8pm (as
    no scans take place outside of these hours).
    """

    try:

        # see note about non-linearity
        # in dateAsYearFaction. This
        # could also potentially be non-
        # monotonic because we are not
        # taking into account leap-seconds.
        t          = t.timetuple()
        daysinyear = 366 if calendar.isleap(t.tm_year) else 365
        dayfrac    = ((t.tm_hour - 7)    +
                      (t.tm_min  / 60.0) +
                      (t.tm_sec  / 3600.0)) / 13.0

        return t.tm_year + ((t.tm_yday - 1) + dayfrac) / float(daysinyear)
    except Exception:
        return np.nan


@funpack.formatter('FMRIBImagingTime')
def format_normalisedAcquisitionTime(dtable, column, series):
    """Formats timestamps using :func:`normalisedAcquisitionTime`. """
    return series.apply(normalisedAcquisitionTime)



#!/usr/bin/env python
#
# test_fmrib.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import textwrap as tw

import pandas as pd
import numpy  as np

import funpack.plugins.fmrib as fmrib


from . import tempdir, gen_DataTableFromDataFrame

TEST_DATES = tw.dedent("""
    2016-01-01
    2016-12-31
    2017-01-01
    2017-12-31
    2018-01-01
    2018-12-31
    """).strip().split('\n')

EXPECT_DATES = [
    2016.0, 2016.9972677595629,
    2017.0, 2017.9972602739726,
    2018.0, 2018.9972602739726]


TEST_TIMES = tw.dedent("""
    2016-01-01T07:00:00+0
    2016-01-01T19:59:59+0
    2016-12-31T07:00:00+0
    2016-12-31T19:59:59+0
    2017-01-01T07:00:00+0
    2017-01-01T19:59:59+0
    2017-12-31T07:00:00+0
    2017-12-31T19:59:59+0
    2018-01-01T07:00:00+0
    2018-01-01T19:59:59+0
    2018-12-31T07:00:00+0
    2018-12-31T19:59:59+0
    """).strip().split('\n')


EXPECT_TIMES = [
    2016.0, 2016.002732182056,  2016.9972677595629, 2016.9999999416189,
    2017.0, 2017.0027396674861, 2017.9972602739726, 2017.9999999414588,
    2018.0, 2018.0027396674861, 2018.9972602739726, 2018.9999999414588]


def test_columns_FMRIBImaging():
    fmrib.columns_FMRIBImaging('abc')


def test_load_FMRIBImaging():

    data = tw.dedent("""
    1234567  20150311  133817.240000  5.01  2  0
    1847293  20150428  130036.235000  5.01  1  1
    6821357  20151004  175803.285000  5.02  0  2
    """).strip()

    expect_index = [1234567, 1847293, 6821357]
    expect_cols  = {
        'acq_time'         : [pd.to_datetime('2015-03-11T13:38:17.24+0'),
                              pd.to_datetime('2015-04-28T13:00:36.235+0'),
                              pd.to_datetime('2015-10-04T17:58:03.285+0')],
        'acq_phase'        : [5.01, 5.01, 5.02],
        'processing_phase' : [2, 1, 0],
        'flipped_SWI'      : [0, 1, 2]
    }

    with tempdir():

        with open('data.tsv', 'wt') as f:
            f.write(data)

        result = fmrib.load_FMRIBImaging('data.tsv')

        assert np.all(result.index == expect_index)
        assert np.all(sorted(result.columns) == sorted(expect_cols.keys()))

        for colname in expect_cols.keys():
            col = pd.Series(expect_cols[colname], index=expect_index)
            assert (col == result[colname]).all()


def test_dateAsYearFraction():

    dates  = TEST_DATES
    expect = EXPECT_DATES

    dates  = [pd.to_datetime(d) for d in dates]
    result = [fmrib.dateAsYearFraction(d) for d in dates]

    assert np.all(np.isclose(expect, result))
    assert np.isnan(fmrib.dateAsYearFraction('notadate'))


def test_normalisedAcquisitionTime():

    times  = TEST_TIMES
    expect = EXPECT_TIMES

    times  = [pd.to_datetime(t) for t in times]
    result = [fmrib.normalisedAcquisitionTime(t) for t in times]

    assert np.all(np.isclose(expect, result))
    assert np.isnan(fmrib.normalisedAcquisitionTime('notatime'))


def test_format_funcs():

    dates   = TEST_DATES
    times   = TEST_TIMES[:len(dates)]
    dexpect = EXPECT_DATES
    texpect = EXPECT_TIMES[:len(dates)]

    dates = [pd.to_datetime(d) for d in dates]
    times = [pd.to_datetime(t) for t in times]


    df = pd.DataFrame()
    df['eid'] = range(1, len(dates) + 1)
    df.set_index('eid', inplace=True)
    df['1-0.0'] = dates
    df['2-0.0'] = times

    dt = gen_DataTableFromDataFrame(df)

    dresult = fmrib.format_dateAsYearFraction(
        dt, '1-0.0', dt[:, '1-0.0'])
    tresult = fmrib.format_normalisedAcquisitionTime(
        dt, '2-0.0', dt[:, '2-0.0'])

    assert (dresult == dexpect).all()
    assert (tresult == texpect).all()
